"""Read a Delwaq model generated from a Ribasim model and inject the results back to Ribasim."""

from pathlib import Path

import ribasim
from ribasim.utils import MissingOptionalModule, _concat

try:
    import xugrid as xu
except ImportError:
    xu = MissingOptionalModule("xugrid", "delwaq")

delwaq_dir = Path(__file__).parent
repo_dir = delwaq_dir.parents[1]
output_folder = delwaq_dir / "model"


def parse(
    toml_path: Path, graph, substances, output_folder=output_folder
) -> ribasim.Model:
    model = ribasim.Model.read(toml_path)

    # Output of Delwaq
    ug = xu.open_dataset(output_folder / "delwaq_map.nc")

    mapping = dict(graph.nodes(data="id"))
    # Continuity is a (default) tracer representing the mass balance
    substances.add("Continuity")

    dfs = []
    for substance in substances:
        df = ug[f"ribasim_{substance}"].to_dataframe().reset_index()
        df.rename(
            columns={
                "ribasim_nNodes": "node_id",
                "nTimesDlwq": "time",
                f"ribasim_{substance}": "concentration",
            },
            inplace=True,
        )
        df["substance"] = substance
        df.drop(columns=["ribasim_node_x", "ribasim_node_y"], inplace=True)
        # Map the node_id (logical index) to the original node_id
        # TODO Check if this is correct
        df.node_id += 1
        df.node_id = df.node_id.map(mapping)

        dfs.append(df)

    df = _concat(dfs).reset_index(drop=True)
    df.sort_values(["time", "node_id"], inplace=True)

    model.basin.concentration_external = df
    df.to_feather(toml_path.parent / "results" / "basin_concentration_external.arrow")

    return model



from contextlib import closing
from pathlib import Path
from sqlite3 import Connection, connect


def esc_id(identifier: str) -> str:
    """Escape SQLite identifiers."""
    identifier = identifier.replace('"', '""')
    return f'"{identifier}"'


def exists(connection: Connection, name: str) -> bool:
    """Check if a table exists in a SQLite database."""
    with closing(connection.cursor()) as cursor:
        cursor.execute(
            "SELECT name FROM sqlite_master WHERE type='table' AND name=?", (name,)
        )
        result = cursor.fetchone()
    return result is not None


def _set_gpkg_attribute_table(connection: Connection, table: str) -> None:
    # Set geopackage attribute table
    with closing(connection.cursor()) as cursor:
        sql = "INSERT OR REPLACE INTO gpkg_contents (table_name, data_type, identifier) VALUES (?, ?, ?)"
        cursor.execute(sql, (table, "attributes", table))
    connection.commit()


CREATE_TABLE_SQL = """
CREATE TABLE IF NOT EXISTS ribasim_metadata (
    key TEXT PRIMARY KEY,
    value TEXT
);
"""


def _get_db_schema_version(db_path: Path) -> int:
    """
    Get the schema version of the database.

    For older models, the version is assumed to be zero,
    which is smaller than the initial schema version of the database.
    """
    with closing(connect(db_path)) as connection:
        if not exists(connection, "ribasim_metadata"):
            return 0
        with closing(connection.cursor()) as cursor:
            cursor.execute(
                "SELECT value FROM ribasim_metadata WHERE key='schema_version'"
            )
            return int(cursor.fetchone()[0])


def _set_db_schema_version(db_path: Path, version: int = 1) -> None:
    with closing(connect(db_path)) as connection:
        if not exists(connection, "metadata"):
            with closing(connection.cursor()) as cursor:
                cursor.execute(CREATE_TABLE_SQL)
                cursor.execute(
                    "INSERT OR REPLACE INTO ribasim_metadata (key, value) VALUES ('schema_version', ?)",
                    (version,),
                )
            _set_gpkg_attribute_table(connection, "ribasim_metadata")
            connection.commit()



"""Setup a Delwaq model from a Ribasim model and results."""

import argparse
import csv
import logging
import shutil
from datetime import timedelta
from pathlib import Path

from ribasim import nodes
from ribasim.utils import MissingOptionalModule, _concat, _pascal_to_snake

try:
    import networkx as nx
except ImportError:
    nx = MissingOptionalModule("networkx", "delwaq")

import numpy as np
import pandas as pd

try:
    import jinja2
except ImportError:
    jinja2 = MissingOptionalModule("jinja2", "delwaq")  # type: ignore

import ribasim
from ribasim.delwaq.util import (
    strfdelta,
    ugrid,
    write_flows,
    write_pointer,
    write_volumes,
)

logger = logging.getLogger(__name__)
delwaq_dir = Path(__file__).parent
output_path = delwaq_dir / "model"

env = jinja2.Environment(
    autoescape=True, loader=jinja2.FileSystemLoader(delwaq_dir / "template")
)

# Add evaporation links, so mass balance is correct
# To simulate salt increase due to evaporation, set to False
USE_EVAP = True


def _boundary_name(id, type):
    # Delwaq has a limit of 12 characters for the boundary name
    return type[:9] + "_" + str(id)


def _quote(value):
    return f"'{value}'"


def _make_boundary(data, boundary_type):
    """
    Create a Delwaq boundary definition with the given data and boundary type.

    Pivot our data from long to wide format, and convert the time to a string.
    Specifically, we go from a table:
        `node_id, substance, time, concentration`
    to
        ```
        ITEM 'Drainage_6'
        CONCENTRATIONS 'Cl' 'Tracer'
        ABSOLUTE TIME
        LINEAR DATA 'Cl' 'Tracer'
        '2020/01/01-00:00:00' 0.0  1.0
        '2020/01/02-00:00:00' 1.0 -999
        ```
    """
    bid = _boundary_name(data.node_id.iloc[0], boundary_type)
    piv = (
        data.pivot_table(
            index="time", columns="substance", values="concentration", fill_value=-999
        )
        .reset_index()
        .reset_index(drop=True)
    )
    # Convert Arrow time to Numpy to avoid needing tzdata somehow
    piv.time = piv.time.astype("datetime64[ns]").dt.strftime("%Y/%m/%d-%H:%M:%S")
    boundary = {
        "name": bid,
        "substances": list(map(_quote, piv.columns[1:])),
        "df": piv.to_string(formatters={"time": _quote}, header=False, index=False),
    }
    substances = data.substance.unique()
    return boundary, substances


def _setup_graph(nodes, link, evaporate_mass=True):
    G = nx.DiGraph()

    assert nodes.df is not None
    for row in nodes.df.itertuples():
        if row.node_type not in ribasim.geometry.link.SPATIALCONTROLNODETYPES:
            G.add_node(
                row.Index,
                type=row.node_type,
                id=row.Index,
                x=row.geometry.x,
                y=row.geometry.y,
                pos=(row.geometry.x, row.geometry.y),
            )
    assert link.df is not None
    for row in link.df.itertuples():
        if row.link_type == "flow":
            G.add_edge(
                row.from_node_id,
                row.to_node_id,
                id=[row.Index],
                duplicate=None,
            )

    # Simplify network, only keeping Basins and Boundaries.
    # We find an unwanted node, remove it,
    # and merge the flow links to/from the node.
    remove_nodes = []
    for node_id, out in G.succ.items():
        if G.nodes[node_id]["type"] not in [
            "Basin",
            "Terminal",
            "LevelBoundary",
            "FlowBoundary",
            "UserDemand",
        ]:
            inneighbor_ids = G.pred[node_id]
            assert len(inneighbor_ids) == 1
            inneighbor_id = list(inneighbor_ids)[0]
            remove_nodes.append(node_id)

            for outneighbor_id in out.keys():
                if outneighbor_id in remove_nodes:
                    logger.debug("Not making link to removed node.")
                    continue
                link = (inneighbor_id, outneighbor_id)
                link_id = G.get_edge_data(node_id, outneighbor_id)["id"][0]
                if G.has_edge(*link):
                    data = G.get_edge_data(*link)
                    data["id"].append(link_id)
                else:
                    G.add_edge(*link, id=[link_id])

    iso = nx.number_of_isolates(G)
    if iso > 0:
        logger.debug(f"Found {iso} isolated nodes in the network.")
        remove_nodes.extend(list(nx.isolates(G)))

    for node_id in remove_nodes:
        G.remove_node(node_id)

    # Due to the simplification, we can end up with cycles of length 2.
    # This happens when a UserDemand is connected to and from a Basin,
    # but can also happen in other cases (rivers with a outlet and pump),
    # for which we do nothing. We merge these UserDemand cycles links to
    # a single link, and later merge the flows.
    merge_links = []
    for loop in nx.simple_cycles(G):
        if len(loop) == 2:
            if (
                G.nodes[loop[0]]["type"] != "UserDemand"
                and G.nodes[loop[1]]["type"] != "UserDemand"
            ):
                logger.debug("Found cycle that is not a UserDemand.")
            else:
                link_ids = G.edges[loop]["id"]
                G.edges[reversed(loop)]["id"].extend(link_ids)
                merge_links.extend(link_ids)
                G.remove_link(*loop)

    # Remove boundary to boundary links
    remove_double_links = []
    for x in G.edges(data=True):
        a, b, d = x
        if G.nodes[a]["type"] == "Terminal" and G.nodes[b]["type"] == "UserDemand":
            logger.debug("Removing link between Terminal and UserDemand")
            remove_double_links.append(a)
        elif G.nodes[a]["type"] == "UserDemand" and G.nodes[b]["type"] == "Terminal":
            remove_double_links.append(b)
            logger.debug("Removing link between UserDemand and Terminal")

    for node_id in remove_double_links:
        G.remove_node(node_id)

    # Relabel the nodes as consecutive integers for Delwaq
    # Note that the node["id"] is the original node_id
    basin_id = 0
    boundary_id = 0
    node_mapping = {}
    basin_mapping: dict[int, int] = {}
    for node_id, node in G.nodes.items():
        if node["type"] == "Basin":
            basin_id += 1
            node_mapping[node_id] = basin_id
            basin_mapping[node["id"]] = basin_id
        elif node["type"] in [
            "Terminal",
            "UserDemand",
            "LevelBoundary",
            "FlowBoundary",
        ]:
            boundary_id -= 1
            node_mapping[node_id] = boundary_id
        else:
            raise ValueError(f"Found unexpected node {node_id} in delwaq graph.")

    nx.relabel_nodes(G, node_mapping, copy=False)

    # Add basin boundaries
    for node_id, node in list(G.nodes(data=True)):
        if node["type"] == "Basin":
            boundary_id -= 1
            G.add_node(
                boundary_id,
                type="Drainage",
                id=node["id"],
                pos=(node["pos"][0] - 0.5, node["pos"][1] + 0.5),
            )
            G.add_edge(
                boundary_id,
                node_id,
                key=link_id,
                id=[-1],
                boundary=(node["id"], "drainage"),
            )

            boundary_id -= 1
            G.add_node(
                boundary_id,
                type="Precipitation",
                id=node["id"],
                pos=(node["pos"][0] + 0, node["pos"][1] + 0.5),
            )
            G.add_edge(
                boundary_id,
                node_id,
                key=link_id,
                id=[-1],
                boundary=(node["id"], "precipitation"),
            )

            if evaporate_mass:
                boundary_id -= 1
                G.add_node(
                    boundary_id,
                    type="Evaporation",
                    id=node["id"],
                    pos=(node["pos"][0] + 0.5, node["pos"][1] + 0.5),
                )
                G.add_edge(
                    node_id,
                    boundary_id,
                    key=link_id,
                    id=[-1],
                    boundary=(node["id"], "evaporation"),
                )

    # Setup link mapping
    link_mapping = {}
    for i, (a, b, d) in enumerate(G.edges(data=True)):
        for link_id in d["id"]:
            link_mapping[link_id] = i

    assert len(basin_mapping) == basin_id

    return G, merge_links, node_mapping, link_mapping, basin_mapping


def _setup_boundaries(model):
    boundaries = []
    substances = set()

    if model.level_boundary.concentration.df is not None:
        for _, rows in model.level_boundary.concentration.df.groupby(["node_id"]):
            boundary, substance = _make_boundary(rows, "LevelBoundary")
            boundaries.append(boundary)
            substances.update(substance)

    if model.flow_boundary.concentration.df is not None:
        for _, rows in model.flow_boundary.concentration.df.groupby("node_id"):
            boundary, substance = _make_boundary(rows, "FlowBoundary")
            boundaries.append(boundary)
            substances.update(substance)

    if model.basin.concentration.df is not None:
        for _, rows in model.basin.concentration.df.groupby(["node_id"]):
            for boundary_type in ("Drainage", "Precipitation"):
                nrows = rows.rename(columns={boundary_type.lower(): "concentration"})
                boundary, substance = _make_boundary(nrows, boundary_type)
                boundaries.append(boundary)
                substances.update(substance)

    return boundaries, substances


def generate(
    toml_path: Path,
    output_path: Path = output_path,
) -> tuple[nx.DiGraph, set[str]]:
    """Generate a Delwaq model from a Ribasim model and results."""
    # Read in model and results
    model = ribasim.Model.read(toml_path)
    results_folder = toml_path.parent / model.results_dir
    evaporate_mass = model.solver.evaporate_mass

    basins = pd.read_feather(
        toml_path.parent / results_folder / "basin.arrow", dtype_backend="pyarrow"
    )
    flows = pd.read_feather(
        toml_path.parent / results_folder / "flow.arrow", dtype_backend="pyarrow"
    )

    output_path.mkdir(exist_ok=True)

    # Setup flow network
    G, merge_links, node_mapping, link_mapping, basin_mapping = _setup_graph(
        model.node_table(), model.link, evaporate_mass=evaporate_mass
    )

    # Plot
    # plt.figure(figsize=(18, 18))
    # nx.draw(
    #     G,
    #     pos={k: v["pos"] for k, v in G.nodes(data=True)},
    #     with_labels=True,
    #     labels={k: v["id"] for k, v in G.nodes(data=True)},
    # )
    # plt.savefig("after_relabeling.png", dpi=300)

    # Setup metadata
    if model.solver.saveat == 0 or np.isposinf(model.solver.saveat):
        raise ValueError("Unsupported saveat, must be positive and finite.")
    else:
        timestep = timedelta(seconds=model.solver.saveat)

    # Write topology to delwaq pointer file
    pointer = pd.DataFrame(G.edges(), columns=["from_node_id", "to_node_id"])
    write_pointer(output_path / "ribasim.poi", pointer)
    pointer["riba_link_id"] = [e[2] for e in G.edges.data("id")]
    pointer["riba_from_node_id"] = pointer["from_node_id"].map(
        {v: k for k, v in node_mapping.items()}
    )
    pointer["riba_to_node_id"] = pointer["to_node_id"].map(
        {v: k for k, v in node_mapping.items()}
    )
    pointer.to_csv(output_path / "network.csv", index=False)  # not needed

    total_segments = len(basin_mapping)
    total_exchanges = len(pointer)

    # Write attributes template
    template = env.get_template("delwaq.atr.j2")
    with open(output_path / "ribasim.atr", mode="w") as f:
        f.write(
            template.render(
                nsegments=total_segments,
            )
        )

    # Generate mesh and write to NetCDF
    uds = ugrid(G)
    uds.ugrid.to_netcdf(output_path / "ribasim.nc")

    # Generate area and flows
    # File format is int32, float32 based
    # Time is internal clock, not real time!
    flows.time = (flows.time - flows.time[0]).dt.total_seconds().astype("int32")
    basins.time = (basins.time - basins.time[0]).dt.total_seconds().astype("int32")

    # Invert flows for half-link of cycles so later summing is correct
    m = flows.link_id.isin(merge_links)
    flows.loc[m, "flow_rate"] = flows.loc[m, "flow_rate"] * -1

    # Map link_id to the new link_id and merge any duplicate flows
    flows["riba_link_id"] = flows["link_id"]
    flows["link_id"] = flows["link_id"].map(link_mapping)
    flows.dropna(subset=["link_id"], inplace=True)
    flows["link_id"] = flows["link_id"].astype("int32")
    nflows = flows.copy()
    nflows = flows.groupby(["time", "link_id"]).sum().reset_index()
    nflows.drop(
        columns=["from_node_id", "to_node_id"],
        inplace=True,
    )

    # Add basin boundaries to flows
    for link_id, (a, b, (node_id, boundary_type)) in enumerate(
        G.edges(data="boundary", default=(None, None))
    ):
        if boundary_type is None:
            continue
        df = basins[basins.node_id == node_id][["time", boundary_type]].rename(
            columns={boundary_type: "flow_rate"}
        )
        df["link_id"] = link_id
        nflows = _concat([nflows, df], ignore_index=True)

    # Save flows to Delwaq format
    nflows.sort_values(by=["time", "link_id"], inplace=True)
    nflows.to_csv(output_path / "flows.csv", index=False)  # not needed
    nflows.drop(
        columns=["link_id", "riba_link_id"],
        inplace=True,
    )
    write_flows(output_path / "ribasim.flo", nflows, timestep)
    write_flows(
        output_path / "ribasim.are", nflows, timestep
    )  # same as flow, so area becomes 1

    # Write volumes to Delwaq format
    basins.drop(columns=["level"], inplace=True)
    volumes = basins[["time", "node_id", "storage"]]
    volumes["riba_node_id"] = volumes["node_id"]
    volumes.loc[:, "node_id"] = (
        volumes["node_id"].map(basin_mapping).astype(pd.Int32Dtype())
    )
    volumes = volumes.sort_values(by=["time", "node_id"])
    volumes.to_csv(output_path / "volumes.csv", index=False)  # not needed
    volumes.drop(columns=["node_id", "riba_node_id"], inplace=True)
    write_volumes(output_path / "ribasim.vol", volumes, timestep)
    write_volumes(
        output_path / "ribasim.vel", volumes, timestep
    )  # same as volume, so vel becomes 1

    # Length file
    lengths = nflows.copy()
    lengths.flow_rate = 1
    lengths.iloc[np.repeat(np.arange(len(lengths)), 2)]
    write_flows(output_path / "ribasim.len", lengths, timestep)

    # Find all boundary substances and concentrations
    boundaries, substances = _setup_boundaries(model)

    # Write boundary data with substances and concentrations
    template = env.get_template("B5_bounddata.inc.j2")
    with open(output_path / "B5_bounddata.inc", mode="w") as f:
        f.write(
            template.render(
                states=[],  # no states yet
                boundaries=boundaries,
            )
        )

    # Setup initial basin concentrations
    defaults = {
        "Continuity": 1.0,
        "Initial": 1.0,
        "LevelBoundary": 0.0,
        "FlowBoundary": 0.0,
        "Terminal": 0.0,
        "UserDemand": 0.0,
        "Precipitation": 0.0,
        "Drainage": 0.0,
    }
    substances.update(defaults.keys())

    # Add user defined substances
    if model.basin.concentration_state.df is not None:
        initial = model.basin.concentration_state.df
        substances.update(initial.substance.unique())

    # Make a wide table with the initial default concentrations
    # using zero for all user defined substances
    icdf = pd.DataFrame(
        data={
            substance: [defaults.get(substance, 0.0)] * len(basin_mapping)
            for substance in sorted(substances)
        },
        index=list(basin_mapping.values()),
    )

    # Override default concentrations with the user defined values
    if model.basin.concentration_state.df is not None:
        for _, row in initial.iterrows():
            icdf.loc[basin_mapping[row.node_id], row.substance] = row.concentration

    # Add comment with original Basin ID
    reverse_node_mapping = {v: k for k, v in node_mapping.items()}
    icdf["comment"] = [f"; {reverse_node_mapping[k]}" for k in icdf.index]

    initial_concentrations = icdf.to_string(header=False, index=False)

    # Write boundary list, ordered by bid to map the unique boundary names
    # to the links described in the pointer file.
    bnd = pointer.copy()
    bnd["bid"] = np.minimum(bnd["from_node_id"], bnd["to_node_id"])
    bnd = bnd[bnd["bid"] < 0]
    bnd.sort_values(by="bid", ascending=False, inplace=True)
    bnd["node_type"] = [G.nodes(data="type")[bid] for bid in bnd["bid"]]
    bnd["node_id"] = [G.nodes(data="id")[bid] for bid in bnd["bid"]]
    bnd["fid"] = list(map(_boundary_name, bnd["node_id"], bnd["node_type"]))
    bnd["comment"] = ""
    bnd.to_csv(output_path / "bndlist.csv", index=False)
    bnd = bnd[["fid", "comment", "node_type"]]
    bnd.drop_duplicates(subset="fid", inplace=True)
    assert bnd["fid"].is_unique

    bnd.to_csv(
        output_path / "ribasim_bndlist.inc",
        index=False,
        header=False,
        sep=" ",
        quotechar="'",
        quoting=csv.QUOTE_ALL,
    )

    # Setup DIMR configuration for running Delwaq via DIMR
    dimrc = delwaq_dir / "reference/dimr_config.xml"
    shutil.copy(dimrc, output_path / "dimr_config.xml")

    # Write main Delwaq input file
    template = env.get_template("delwaq.inp.j2")
    with open(output_path / "delwaq.inp", mode="w") as f:
        f.write(
            template.render(
                startime=model.starttime,
                endtime=model.endtime - timestep,
                timestep=strfdelta(timestep),
                nsegments=total_segments,
                nexchanges=total_exchanges,
                substances=sorted(substances),
                initial_concentrations=initial_concentrations,
            )
        )

    # Return the graph with original links and the substances
    # so we can parse the results back to the original model
    return G, substances


def add_tracer(model, node_id, tracer_name):
    """Add a tracer to the Delwaq model."""
    n = model.node_table().df.loc[node_id]
    node_type = n.node_type
    if node_type not in [
        "Basin",
        "LevelBoundary",
        "FlowBoundary",
        "UserDemand",
    ]:
        raise ValueError("Can only trace Basins and boundaries")
    snake_node_type = _pascal_to_snake(node_type)
    nt = getattr(model, snake_node_type)

    ct = getattr(nodes, snake_node_type)
    table = ct.Concentration(
        node_id=[node_id],
        time=[model.starttime],
        substance=[tracer_name],
        concentration=[1.0],
    )
    if nt.concentration is None:
        nt.concentration = table
    else:
        nt.concentration = pd.concat([nt.concentration.df, table.df], ignore_index=True)


if __name__ == "__main__":
    # Generate a Delwaq model from the default Ribasim model

    parser = argparse.ArgumentParser(
        description="Generate Delwaq input from Ribasim results."
    )
    parser.add_argument(
        "toml_path", type=Path, help="The path to the Ribasim TOML file."
    )
    parser.add_argument(
        "--output_path",
        type=Path,
        help="The relative path to store the Delwaq model.",
        default="delwaq",
    )
    args = parser.parse_args()

    graph, substances = generate(
        args.toml_path, args.toml_path.parent / args.output_path
    )



from ribasim.input_base import TableModel
from ribasim.schemas import (
    TabulatedRatingCurveStaticSchema,
    TabulatedRatingCurveTimeSchema,
)

__all__ = ["Static", "Time"]


class Static(TableModel[TabulatedRatingCurveStaticSchema]):
    pass


class Time(TableModel[TabulatedRatingCurveTimeSchema]):
    pass



from ribasim.geometry.area import BasinAreaSchema
from ribasim.input_base import SpatialTableModel, TableModel
from ribasim.schemas import (
    BasinConcentrationExternalSchema,
    BasinConcentrationSchema,
    BasinConcentrationStateSchema,
    BasinProfileSchema,
    BasinStateSchema,
    BasinStaticSchema,
    BasinSubgridSchema,
    BasinSubgridTimeSchema,
    BasinTimeSchema,
)

__all__ = [
    "Area",
    "Concentration",
    "Profile",
    "State",
    "Static",
    "Subgrid",
    "SubgridTime",
    "Time",
]


class Static(TableModel[BasinStaticSchema]):
    pass


class Time(TableModel[BasinTimeSchema]):
    pass


class State(TableModel[BasinStateSchema]):
    pass


class Profile(TableModel[BasinProfileSchema]):
    pass


class Subgrid(TableModel[BasinSubgridSchema]):
    pass


class SubgridTime(TableModel[BasinSubgridTimeSchema]):
    pass


class Area(SpatialTableModel[BasinAreaSchema]):
    pass


class Concentration(TableModel[BasinConcentrationSchema]):
    pass


class ConcentrationExternal(TableModel[BasinConcentrationExternalSchema]):
    pass


class ConcentrationState(TableModel[BasinConcentrationStateSchema]):
    pass



import os
from pathlib import Path

import pytest
from ribasim import Model
from ribasim.delwaq import add_tracer, generate, parse, run_delwaq

delwaq_dir = Path(__file__).parent


@pytest.mark.skipif(
    not (
        ("D3D_HOME" in os.environ)
        and (
            delwaq_dir.parents[2] / "generated_testmodels/basic/results/flow.arrow"
        ).is_file()
    ),
    reason="Requires Delwaq to be installed and basic model results.",
)
def test_offline_delwaq_coupling():
    repo_dir = delwaq_dir.parents[2]
    toml_path = repo_dir / "generated_testmodels/basic/ribasim.toml"

    model = Model.read(toml_path)
    add_tracer(model, 11, "Foo")
    add_tracer(model, 15, "Bar")
    model.write(toml_path)

    graph, substances = generate(toml_path)
    run_delwaq()
    model = parse(toml_path, graph, substances)

    df = model.basin.concentration_external.df
    assert df is not None
    assert df.shape[0] > 0
    assert df.node_id.nunique() == 4
    assert sorted(df.substance.unique()) == [
        "Bar",
        "Cl",
        "Continuity",
        "Drainage",
        "FlowBoundary",
        "Foo",
        "Initial",
        "LevelBoundary",
        "Precipitation",
        "Terminal",
        "Tracer",
        "UserDemand",
    ]



from unittest.mock import patch

import pandas as pd
import pytest
import ribasim
from pandas.testing import assert_frame_equal
from pydantic import ValidationError
from ribasim import Model
from ribasim.db_utils import _get_db_schema_version, _set_db_schema_version
from ribasim.migrations import _rename_column
from ribasim.nodes import basin
from ribasim.schemas import BasinProfileSchema
from shapely.geometry import Point


def test_config_inheritance():
    assert BasinProfileSchema.__config__.add_missing_columns
    assert BasinProfileSchema.__config__.coerce


@patch("ribasim.schemas.migrations.nodeschema_migration")
def test_migration(migration, basic, tmp_path):
    toml_path = tmp_path / "basic.toml"
    db_path = tmp_path / "database.gpkg"
    basic.write(toml_path)

    # Migration is not needed on default model
    Model.read(toml_path)
    assert not migration.called

    # Fake old schema that needs migration
    _set_db_schema_version(db_path, 0)
    Model.read(toml_path)
    assert migration.called


def test_model_schema(basic, tmp_path):
    toml_path = tmp_path / "basic.toml"
    db_path = tmp_path / "database.gpkg"
    basic.write(toml_path)

    assert _get_db_schema_version(db_path) == ribasim.__schema_version__
    _set_db_schema_version(db_path, 0)
    assert _get_db_schema_version(db_path) == 0


def test_geometry_validation():
    with pytest.raises(
        ValidationError,
        match="Column 'geometry' failed element-wise validator number 0: <Check is_correct_geometry_type> failure cases",
    ):
        basin.Area(geometry=[Point([1.0, 2.0])])


def test_column_rename():
    df = pd.DataFrame({"edge_type": [1], "link_type": [2]})
    _rename_column(df, "edge_type", "link_type")
    assert_frame_equal(df, pd.DataFrame({"link_type": [1]}))
    df = pd.DataFrame({"edge_type": [1]})
    _rename_column(df, "edge_type", "link_type")
    assert_frame_equal(df, pd.DataFrame({"link_type": [1]}))
    df = pd.DataFrame({"link_type": [2]})
    assert_frame_equal(df, _rename_column(df, "edge_type", "link_type"))



import geopandas as gpd
import pytest
import shapely.geometry as sg
from pydantic import ValidationError
from ribasim.geometry.link import LinkTable, NodeData


@pytest.fixture(scope="session")
def link() -> LinkTable:
    a = (0.0, 0.0)
    b = (0.0, 1.0)
    c = (0.2, 0.5)
    d = (1.0, 1.0)
    geometry = [sg.LineString([a, b, c]), sg.LineString([a, d])]
    df = gpd.GeoDataFrame(
        data={"link_id": [0, 1], "from_node_id": [1, 1], "to_node_id": [2, 3]},
        geometry=geometry,
    )
    df.set_index("link_id", inplace=True)
    link = LinkTable(df=df)
    return link


def test_validation(link):
    assert isinstance(link, LinkTable)

    with pytest.raises(ValidationError):
        df = gpd.GeoDataFrame(
            data={
                "link_id": [0, 1],
                "from_node_id": [1, 1],
                "to_node_id": ["foo", 3],
            },  # None is coerced to 0 without errors
            geometry=[None, None],
        )
        df.set_index("link_id", inplace=True)
        LinkTable(df=df)


def test_link_plot(link):
    link.plot()


def test_link_indexing(link):
    with pytest.raises(NotImplementedError):
        link[1]


def test_invalid_retour_link(basic):
    with pytest.raises(ValueError, match="opposite link already exists"):
        basic.link.add(basic.manning_resistance[2], basic.basin[1])


def test_node_data():
    node = NodeData(node_id=5, node_type="Pump", geometry=sg.Point(0, 0))
    assert repr(node) == "Pump #5"



import pytest
import ribasim
import ribasim_testmodels


# we can't call fixtures directly, so we keep separate versions
@pytest.fixture()
def basic() -> ribasim.Model:
    return ribasim_testmodels.basic_model()


@pytest.fixture()
def basic_arrow() -> ribasim.Model:
    return ribasim_testmodels.basic_arrow_model()


@pytest.fixture()
def basic_transient() -> ribasim.Model:
    return ribasim_testmodels.basic_transient_model()


@pytest.fixture()
def bucket() -> ribasim.Model:
    return ribasim_testmodels.bucket_model()


@pytest.fixture()
def pid_control_equation() -> ribasim.Model:
    return ribasim_testmodels.pid_control_equation_model()


@pytest.fixture()
def tabulated_rating_curve() -> ribasim.Model:
    return ribasim_testmodels.tabulated_rating_curve_model()


@pytest.fixture()
def outlet() -> ribasim.Model:
    return ribasim_testmodels.outlet_model()


@pytest.fixture()
def backwater() -> ribasim.Model:
    return ribasim_testmodels.backwater_model()


@pytest.fixture()
def discrete_control_of_pid_control() -> ribasim.Model:
    return ribasim_testmodels.discrete_control_of_pid_control_model()


@pytest.fixture()
def level_range() -> ribasim.Model:
    return ribasim_testmodels.level_range_model()


@pytest.fixture()
def trivial() -> ribasim.Model:
    return ribasim_testmodels.trivial_model()



import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable


def plot_fraction(
    model,
    node_id,
    tracers=[
        "LevelBoundary",
        "FlowBoundary",
        "UserDemand",
        "Initial",
        "Drainage",
        "Precipitation",
        "Terminal",
    ],
):
    table = model.basin.concentration_external.df
    table = table[table["node_id"] == node_id]
    table = table[table["substance"].isin(tracers)]

    groups = table.groupby("substance")
    stack = {k: v["concentration"].to_numpy() for (k, v) in groups}

    fig, ax = plt.subplots()
    ax.stackplot(
        groups.get_group(tracers[0])["time"],
        stack.values(),
        labels=stack.keys(),
    )
    ax.plot(
        groups.get_group(tracers[0])["time"],
        np.sum(list(stack.values()), axis=0),
        c="black",
        lw=2,
    )
    ax.legend()
    ax.set_title(f"Fraction plot for node {node_id}")
    ax.set_xlabel("Time")
    ax.set_ylabel("Fraction")

    plt.show(fig)


def plot_spatial(model, tracer="Basin", versus=None, limit=0.001):
    table = model.basin.concentration_external.df
    table = table[table["time"] == table["time"].max()]

    if versus is not None:
        vtable = table[table["substance"] == versus]
        vtable.set_index("node_id", inplace=True)
    table = table[table["substance"] == tracer]
    table.set_index("node_id", inplace=True)

    nodes = model.node_table().df
    nodes = nodes[nodes.index.isin(table.index)]

    if versus is None:
        c = table["concentration"][nodes.index]
        alpha = c > limit
    else:
        alpha = (
            table["concentration"][nodes.index] + vtable["concentration"][nodes.index]
        )
        c = table["concentration"][nodes.index] / alpha

    fig, ax = plt.subplots()
    s = ax.scatter(
        nodes.geometry.x,
        nodes.geometry.y,
        c=c,
        clim=(0, 1),
        alpha=alpha,
    )
    dt = table["time"].iloc[0]
    if versus is None:
        ax.set_title(f"Scatter plot for {tracer} tracer at {dt}")
    else:
        ax.set_title(f"Scatter plot for {tracer} vs {versus} tracer at {dt}")

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)

    fig.colorbar(s, cax=cax, orientation="vertical")
    if versus is not None:
        cax.set_ylabel(f"{tracer} fraction vs {versus} fraction")
    else:
        cax.set_ylabel(f"Overall {tracer} fraction")
    ax.set_xlabel("x")
    ax.set_ylabel("y")

    plt.show(fig)



"""Utilities to write Delwaq (binary) input files."""

import os
import platform
import struct
import subprocess
from datetime import timedelta
from pathlib import Path

import numpy as np
import numpy.typing as npt
import pandas as pd

from ribasim.utils import MissingOptionalModule

try:
    import xugrid
except ImportError:
    xugrid = MissingOptionalModule("xugrid")


def strfdelta(tdelta) -> str:
    # dddhhmmss format
    days = tdelta.days
    hours, rem = divmod(tdelta.seconds, 3600)
    minutes, seconds = divmod(rem, 60)
    return f"{days:03d}{hours:02d}{minutes:02d}{seconds:02d}"


def write_pointer(fn: Path | str, data: pd.DataFrame) -> None:
    """Write pointer file for Delwaq.

    The format is a matrix of int32 of links
    with 4 columns: from_node_id, to_node_id, 0, 0

    This saves as column major order for Fortran compatibility.

    Data is a DataFrame with columns from_node_id, to_node_id.
    """
    with open(fn, "wb") as f:
        for a, b in data.to_numpy():
            f.write(struct.pack("<4i", a, b, 0, 0))


def write_lengths(fn: Path | str, data: npt.NDArray[np.float32]) -> None:
    """Write lengths file for Delwaq.

    The format is an int defining time/links (?)
    Followed by a matrix of float32 of 2, n_links
    Defining the length of the half-links.

    This saves as column major order for Fortran compatibility.

    Data is an array of float32.
    """
    with open(fn, "wb") as f:
        f.write(struct.pack("<i", 0))
        f.write(data.astype("float32").tobytes())


def write_volumes(fn: Path | str, data: pd.DataFrame, timestep: timedelta) -> None:
    """Write volumes file for Delwaq.

    The format is an int defining the time
    followed by the volume for each node
    The order should be the same as the nodes in the mesh.

    This saves as column major order for Fortran compatibility.

    Data is a DataFrame with columns time, storage
    """
    with open(fn, "wb") as f:
        for time, group in data.groupby("time"):
            f.write(struct.pack("<i", time))
            f.write(group.storage.to_numpy().astype("float32").tobytes())

        # Delwaq needs an extra timestep after the end
        ntime = time + int(timestep.total_seconds())  # type: ignore
        f.write(struct.pack("<i", ntime))
        f.write(group.storage.to_numpy().astype("float32").tobytes())


def write_flows(fn: Path | str, data: pd.DataFrame, timestep: timedelta) -> None:
    """Write flows file for Delwaq.

    The format is an int defining the time
    followed by the flow for each link
    The order should be the same as the nodes in the pointer.

    This saves as column major order for Fortran compatibility.

    Data is a DataFrame with columns time, flow
    """
    with open(fn, "wb") as f:
        for time, group in data.groupby("time"):
            f.write(struct.pack("<i", time))
            f.write(group.flow_rate.to_numpy().astype("float32").tobytes())

        # Delwaq needs an extra timestep after the end
        ntime = time + int(timestep.total_seconds())  # type: ignore
        f.write(struct.pack("<i", ntime))
        f.write(group.flow_rate.to_numpy().astype("float32").tobytes())


def ugrid(G) -> xugrid.UgridDataset:
    # TODO Deduplicate with ribasim.Model.to_xugrid
    link_df = pd.DataFrame(G.edges(), columns=["from_node_id", "to_node_id"])
    node_df = pd.DataFrame(G.nodes(), columns=["node_id"])
    node_df["x"] = [i[1] for i in G.nodes(data="x")]
    node_df["y"] = [i[1] for i in G.nodes(data="y")]
    node_df = node_df[node_df.node_id > 0].reset_index(drop=True)
    node_df.set_index("node_id", drop=False, inplace=True)
    node_df.sort_index(inplace=True)
    link_df = link_df[
        link_df.from_node_id.isin(node_df.node_id)
        & link_df.to_node_id.isin(node_df.node_id)
    ].reset_index(drop=True)

    node_id = node_df.node_id.to_numpy()
    link_id = link_df.index.to_numpy()
    from_node_id = link_df.from_node_id.to_numpy()
    to_node_id = link_df.to_node_id.to_numpy()

    # from node_id to the node_dim index
    node_lookup = pd.Series(
        index=node_id,
        data=node_id.argsort().astype(np.int32),
        name="node_index",
    )

    grid = xugrid.Ugrid1d(
        node_x=node_df.x,
        node_y=node_df.y,
        fill_value=-1,
        edge_node_connectivity=np.column_stack(
            (
                node_lookup[from_node_id],
                node_lookup[to_node_id],
            )
        ),
        name="ribasim",
    )

    link_dim = grid.edge_dimension
    node_dim = grid.node_dimension

    uds = xugrid.UgridDataset(None, grid)
    uds = uds.assign_coords(node_id=(node_dim, node_id))
    uds = uds.assign_coords(link_id=(link_dim, link_id))
    uds = uds.assign_coords(from_node_id=(link_dim, from_node_id))
    uds = uds.assign_coords(to_node_id=(link_dim, to_node_id))

    return uds


def run_delwaq() -> None:
    d3d_home = os.environ.get("D3D_HOME")
    if d3d_home is None:
        raise ValueError("D3D_HOME is not set.")
    else:
        pd3d_home = Path(d3d_home)
    binfolder = (pd3d_home / "bin").absolute()
    folder = Path(__file__).parent
    inp_path = folder / "model" / "delwaq.inp"
    system = platform.system()
    if system == "Windows":
        # run_delwaq.bat prepends working directory to the inp file
        subprocess.run(
            [binfolder / "run_delwaq.bat", "delwaq.inp"],
            cwd=(folder / "model").absolute(),
            check=True,
        )
    elif system == "Linux":
        subprocess.run([binfolder / "run_delwaq.sh", inp_path.absolute()], check=True)
    else:
        raise OSError(f"No support for running Delwaq automatically on {system}.")



from typing import Any

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pandera as pa
from matplotlib.patches import Patch
from pandera.dtypes import Int32
from pandera.typing import Index, Series
from pandera.typing.geopandas import GeoSeries
from shapely.geometry import Point

from ribasim.input_base import SpatialTableModel

from .base import _GeoBaseSchema

__all__ = ("NodeTable",)


class NodeSchema(_GeoBaseSchema):
    node_id: Index[Int32] = pa.Field(default=0, ge=0, check_name=True)
    name: Series[str] = pa.Field(default="")
    node_type: Series[str] = pa.Field(default="")
    subnetwork_id: Series[pd.Int32Dtype] = pa.Field(
        default=pd.NA, nullable=True, coerce=True
    )
    source_priority: Series[pd.Int32Dtype] = pa.Field(
        default=pd.NA, nullable=True, coerce=True
    )
    cyclic_time: Series[bool] = pa.Field(default=False)
    geometry: GeoSeries[Point] = pa.Field(default=None, nullable=True)

    @classmethod
    def _index_name(self) -> str:
        return "node_id"


class NodeTable(SpatialTableModel[NodeSchema]):
    """The Ribasim nodes as Point geometries."""

    def filter(self, nodetype: str):
        """Filter the node table based on the node type."""
        if self.df is not None:
            mask = self.df[self.df["node_type"] != nodetype].index
            self.df.drop(mask, inplace=True)

    def plot_allocation_networks(self, ax=None, zorder=None) -> Any:
        if ax is None:
            _, ax = plt.subplots()
            ax.axis("off")

        COLOR_SUBNETWORK = "black"
        COLOR_MAIN_NETWORK = "blue"
        ALPHA = 0.25

        contains_main_network = False
        contains_subnetworks = False
        assert self.df is not None

        for subnetwork_id, df_subnetwork in self.df.groupby("subnetwork_id"):
            if subnetwork_id is None:
                continue
            elif subnetwork_id == 1:
                contains_main_network = True
                color = COLOR_MAIN_NETWORK
            else:
                contains_subnetworks = True
                color = COLOR_SUBNETWORK

            hull = gpd.GeoDataFrame(
                geometry=[df_subnetwork.geometry.unary_union.convex_hull]
            )
            hull.plot(ax=ax, color=color, alpha=ALPHA, zorder=zorder)

        handles = []
        labels = []

        if contains_main_network:
            handles.append(Patch(facecolor=COLOR_MAIN_NETWORK, alpha=ALPHA))
            labels.append("Main network")
        if contains_subnetworks:
            handles.append(Patch(facecolor=COLOR_SUBNETWORK, alpha=ALPHA))
            labels.append("Subnetwork")

        return handles, labels

    def plot(self, ax=None, zorder=None) -> Any:
        """
        Plot the nodes. Each node type is given a separate marker.

        Parameters
        ----------
        ax : Optional
            The axis on which the nodes will be plotted.

        Returns
        -------
        None
        """
        if ax is None:
            _, ax = plt.subplots()
            ax.axis("off")

        MARKERS = {
            "Basin": "o",
            "ContinuousControl": "*",
            "DiscreteControl": "*",
            "FlowBoundary": "h",
            "FlowDemand": "h",
            "LevelBoundary": "o",
            "LevelDemand": "o",
            "LinearResistance": "^",
            "ManningResistance": "D",
            "Outlet": "h",
            "PidControl": "x",
            "Pump": "h",
            "TabulatedRatingCurve": "D",
            "Terminal": "s",
            "UserDemand": "s",
            "": "o",
        }

        COLORS = {
            "Basin": "b",
            "ContinuousControl": "0.5",
            "DiscreteControl": "k",
            "FlowBoundary": "m",
            "FlowDemand": "r",
            "LevelBoundary": "g",
            "LevelDemand": "k",
            "LinearResistance": "g",
            "ManningResistance": "r",
            "Outlet": "g",
            "PidControl": "k",
            "Pump": "0.5",  # grayscale level
            "TabulatedRatingCurve": "g",
            "Terminal": "m",
            "UserDemand": "g",
            "": "k",
        }
        if self.df is None:
            return

        for nodetype, df in self.df.groupby("node_type"):
            assert isinstance(nodetype, str)
            marker = MARKERS[nodetype]
            color = COLORS[nodetype]
            ax.scatter(
                df.geometry.x,
                df.geometry.y,
                marker=marker,
                color=color,
                zorder=zorder,
                label=nodetype,
            )

        assert self.df is not None
        geometry = self.df["geometry"]
        for text, xy in zip(self.df.index, np.column_stack((geometry.x, geometry.y))):
            ax.annotate(text=text, xy=xy, xytext=(2.0, 2.0), textcoords="offset points")

        return ax



import operator
import re
import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable, Generator
from contextlib import closing
from contextvars import ContextVar
from pathlib import Path
from sqlite3 import connect
from typing import (
    Any,
    Generic,
    TypeVar,
    cast,
)

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import datacompy

import geopandas as gpd
import numpy as np
import pandas as pd
import pydantic
from pandera.typing import DataFrame
from pandera.typing.geopandas import GeoDataFrame
from pydantic import BaseModel as PydanticBaseModel
from pydantic import (
    ConfigDict,
    DirectoryPath,
    Field,
    PrivateAttr,
    ValidationInfo,
    field_validator,
    model_serializer,
    model_validator,
    validate_call,
)

import ribasim
from ribasim.db_utils import (
    _get_db_schema_version,
    _set_gpkg_attribute_table,
    esc_id,
    exists,
)
from ribasim.schemas import _BaseSchema

from .styles import _add_styles_to_geopackage

__all__ = ("TableModel",)

delimiter = " / "
node_names_snake_case = [
    "basin",
    "continuous_control",
    "discrete_control",
    "flow_boundary",
    "flow_demand",
    "level_boundary",
    "level_demand",
    "linear_resistance",
    "manning_resistance",
    "outlet",
    "pid_control",
    "pump",
    "tabulated_rating_curve",
    "user_demand",
]

context_file_loading: ContextVar[dict[str, Any]] = ContextVar(
    "file_loading", default={}
)
context_file_writing: ContextVar[dict[str, Any]] = ContextVar(
    "file_writing", default={}
)

TableT = TypeVar("TableT", bound=_BaseSchema)


class BaseModel(PydanticBaseModel):
    """Overrides Pydantic BaseModel to set our own config."""

    model_config = ConfigDict(
        validate_assignment=True,
        validate_default=True,
        populate_by_name=True,
        use_enum_values=True,
        extra="allow",
    )

    @classmethod
    def _fields(cls) -> list[str]:
        """Return the names of the fields contained in the Model."""
        return list(cls.model_fields.keys())

    def model_dump(self, **kwargs) -> dict[str, Any]:
        return super().model_dump(serialize_as_any=True, **kwargs)

    def diff(
        self, other: "BaseModel", ignore_meta: bool = False
    ) -> dict[str, Any] | None:
        """
        Compare two instances of a BaseModel.

        ** Warning: This method is experimental and is likely to change. **

        If they are equal, return None. Otherwise, return a nested dictionary with the differences.
        When the differences are not a DataFrame (like the toml config),
        the dict has self and other as key.
        For DataFrames we return a dict with diff as key, and a datacompy Comparison object.

        When ignore_meta is set to True, the meta_* columns in the DataFrames are ignored.
        Note that in that case the key will still be returned and the value will be None.

        Examples
        --------
        >>> nbasic == basic
        False
        >>> x = nbasic.diff(basic)
        {'basin': {'node': {'diff': <datacompy.core.Compare object at 0x16e5a45c0>},
                'static': {'diff': <datacompy.core.Compare object at 0x16eb90080>}},
        'solver': {'saveat': {'other': 86400.0, 'self': 0.0}}}
        >>> x["basin"]["static"]["diff"].report()
        DataComPy Comparison
        --------------------
        ...
        """
        if not (isinstance(other, self.__class__)):
            raise ValueError(f"Cannot compare {self} with {other}")
        if self == other:
            return None
        data = {}
        for key in self._fields():
            self_attr = getattr(self, key)
            other_attr = getattr(other, key)
            if self_attr == other_attr:
                continue
            if isinstance(self_attr, BaseModel):
                data[key] = self_attr.diff(
                    other_attr,
                    ignore_meta=ignore_meta,
                )
            else:
                data[key] = {"self": self_attr, "other": other_attr}
        return data

    # __eq__ from Pydantic BaseModel itself, edited to remove the comparison of private attrs
    # https://github.com/pydantic/pydantic/blob/ff3789d4cc06ee024b7253b919d3e36748a72829/pydantic/main.py#L1069
    # The MIT License (MIT) | Copyright (c) 2017 to present Pydantic Services Inc. and individual contributors.
    def __eq__(self, other: Any) -> bool:
        if isinstance(other, BaseModel):
            self_type = self.__pydantic_generic_metadata__["origin"] or self.__class__
            other_type = (
                other.__pydantic_generic_metadata__["origin"] or other.__class__
            )

            if not (
                self_type == other_type
                # This comparison has been removed, otherwise we recurse because
                # we store the parent of the model in a private attribute
                # and getattr(self, "__pydantic_private__", None)
                # == getattr(other, "__pydantic_private__", None)
                and self.__pydantic_extra__ == other.__pydantic_extra__
            ):
                return False

            if self.__dict__ == other.__dict__:
                return True

            model_fields = type(self).__pydantic_fields__.keys()
            if (
                self.__dict__.keys() <= model_fields
                and other.__dict__.keys() <= model_fields
            ):
                return False

            getter = (
                operator.itemgetter(*model_fields)
                if model_fields
                else lambda _: pydantic._utils._SENTINEL  # type: ignore
            )
            try:
                return getter(self.__dict__) == getter(other.__dict__)
            except KeyError:
                self_fields_proxy = pydantic._utils.SafeGetItemProxy(self.__dict__)  # type: ignore
                other_fields_proxy = pydantic._utils.SafeGetItemProxy(other.__dict__)  # type: ignore
                return getter(self_fields_proxy) == getter(other_fields_proxy)

        else:
            return NotImplemented


class FileModel(BaseModel, ABC):
    """Base class to represent models with a file representation.

    It therefore always has a `filepath` and if it is given on
    initialization, it will parse that file.

    This class extends the `model_validator` option of Pydantic,
    so when when a Path is given to a field with type `FileModel`,
    it doesn't error, but actually initializes the `FileModel`.

    Attributes
    ----------
        filepath (Path | None):
            The path of this FileModel.
    """

    filepath: Path | None = Field(default=None, exclude=True, repr=False)

    @model_validator(mode="before")
    @classmethod
    def _check_filepath(cls, value: Any) -> Any:
        # Enable initialization with a Path.
        if isinstance(value, dict):
            # Pydantic Model init requires a dict
            filepath = value.get("filepath", None)
            if filepath is not None:
                filepath = Path(filepath)
            data = cls._load(filepath)
            data.update(value)
            return data
        elif isinstance(value, Path | str):
            # Pydantic Model init requires a dict
            data = cls._load(Path(value))
            data["filepath"] = value
            return data
        else:
            return value

    @validate_call
    def set_filepath(self, filepath: Path) -> None:
        """Set the filepath of this instance.

        Args:
            filepath (Path): The filepath to set.
        """
        # Disable assignment validation, which would
        # otherwise trigger check_filepath() and _load() again.
        self.model_config["validate_assignment"] = False
        self.filepath = filepath
        self.model_config["validate_assignment"] = True

    @classmethod
    @abstractmethod
    def _load(cls, filepath: Path | None) -> dict[str, Any]:
        """Load the data at filepath and returns it as a dictionary.

        If a derived FileModel does not load data from disk, this should
        return an empty dictionary.

        Args
        ----
            filepath (Path): Path to the data to load.

        Returns
        -------
            dict: The data stored at filepath
        """
        raise NotImplementedError()


class TableModel(FileModel, Generic[TableT]):
    df: DataFrame[TableT] | None = Field(default=None, exclude=True, repr=False)
    _sort_keys: list[str] = PrivateAttr(default=[])

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, TableModel):
            if self.df is None and other.df is None:
                return True
            if self.df is None or other.df is None:
                return False
            else:
                return self.df.equals(other.df)

        return NotImplemented

    def diff(self, other: BaseModel, ignore_meta=False) -> dict[str, Any] | None:
        # Only diff with other TableModel[TableT] instances
        if not (isinstance(other, self.__class__)):
            raise ValueError(f"Cannot compare {self} with {other}")
        # Both are None
        if self.df is None and other.df is None:
            return None
        # Both are DataFrames
        elif self.df is not None and other.df is not None:
            # Differences might've been in the meta columns
            # so we check if the DataFrames are equal again
            if ignore_meta:
                a = self.df.loc[:, self.columns()]
                b = other.df.loc[:, self.columns()]
                if a.equals(b):
                    return None
            else:
                a = self.df
                b = other.df

            comp = datacompy.Compare(
                a, b, on_index=True, df1_name="self", df2_name="other"
            )
            return {"diff": comp}
        # One of the instances is None
        else:
            return {"self": self.df, "other": other.df}

    @field_validator("df", mode="before")
    @classmethod
    def _check_schema(cls, v: DataFrame[TableT]):
        """Allow only extra columns with `meta_` prefix."""
        if isinstance(v, pd.DataFrame | gpd.GeoDataFrame):
            # On reading from geopackage, migrate the tables when necessary
            db_path = context_file_loading.get().get("database")
            if db_path is not None:
                version = _get_db_schema_version(db_path)
                if version < ribasim.__schema_version__:
                    v = cls.tableschema().migrate(v, version)
            for colname in v.columns:
                if colname not in cls.columns() and not colname.startswith("meta_"):
                    raise ValueError(
                        f"Unrecognized column '{colname}'. Extra columns need a 'meta_' prefix."
                    )
        return v

    @model_serializer
    def _set_model(self) -> "str | None":
        return str(self.filepath.name) if self.filepath is not None else None

    @classmethod
    def tablename(cls) -> str:
        """Retrieve tablename based on attached Schema.

        NodeSchema -> Schema
        TabularRatingCurveStaticSchema -> TabularRatingCurve / Static
        """
        cls_string = str(cls.tableschema())
        names: list[str] = re.sub("([A-Z]+)", r" \1", cls_string).split()[:-1]
        names_lowered = [name.lower() for name in names]
        if len(names) == 1:
            return names[0]
        else:
            for n in range(1, len(names_lowered) + 1):
                node_name_snake_case = "_".join(names_lowered[:n])
                if node_name_snake_case in node_names_snake_case:
                    node_name = "".join(names[:n])
                    table_name = "_".join(names_lowered[n:])
                    return node_name + delimiter + table_name
            raise ValueError(f"Found no known node name in {cls_string}")

    @model_validator(mode="before")
    @classmethod
    def _check_dataframe(cls, value: Any) -> Any:
        # Enable initialization with a Dict.
        if isinstance(value, dict) and len(value) > 0 and "df" not in value:
            value = DataFrame(dict(**value))

        # Enable initialization with a DataFrame.
        if isinstance(value, pd.DataFrame | gpd.GeoDataFrame):
            value = {"df": value}

        return value

    def _node_ids(self) -> set[int]:
        node_ids: set[int] = set()
        if self.df is not None and "node_id" in self.df.columns:
            node_ids.update(self.df["node_id"])

        return node_ids

    @classmethod
    def _load(cls, filepath: Path | None) -> dict[str, Any]:
        db = context_file_loading.get().get("database")
        if filepath is not None and db is not None:
            adf = cls._from_arrow(filepath)
            # TODO Store filepath?
            return {"df": adf}
        elif db is not None:
            ddf = cls._from_db(db, cls.tablename())
            return {"df": ddf}
        else:
            return {}

    def _save(self, directory: DirectoryPath, input_dir: DirectoryPath) -> None:
        # TODO directory could be used to save an arrow file
        db_path = context_file_writing.get().get("database")
        self.sort()
        if self.filepath is not None:
            self._write_arrow(self.filepath, directory, input_dir)
        elif db_path is not None:
            self._write_geopackage(db_path)

    def _write_geopackage(self, temp_path: Path) -> None:
        """
        Write the contents of the input to a database.

        Parameters
        ----------
        connection : Connection
            SQLite connection to the database.
        """
        assert self.df is not None
        table = self.tablename()

        with closing(connect(temp_path)) as connection:
            self.df.to_sql(
                table,
                connection,
                index=True,
                if_exists="replace",
                dtype={"fid": "INTEGER PRIMARY KEY AUTOINCREMENT"},
            )

            _set_gpkg_attribute_table(connection, table)
            # Set geopackage attribute table

    def _write_arrow(self, filepath: Path, directory: Path, input_dir: Path) -> None:
        """Write the contents of the input to a an arrow file."""
        assert self.df is not None
        path = directory / input_dir / filepath
        path.parent.mkdir(parents=True, exist_ok=True)
        self.df.to_feather(
            path,
            compression="zstd",
            compression_level=6,
        )

    @classmethod
    def _from_db(cls, path: Path, table: str) -> pd.DataFrame | None:
        with closing(connect(path)) as connection:
            if exists(connection, table):
                query = f"select * from {esc_id(table)}"
                df = pd.read_sql_query(
                    query,
                    connection,
                    # we store TIMESTAMP in SQLite like "2025-05-29 14:16:00"
                    # see https://www.sqlite.org/lang_datefunc.html
                    parse_dates={"time": {"format": "ISO8601"}},
                    dtype_backend="pyarrow",
                )
                df.set_index("fid", inplace=True)
            else:
                df = None

            return df

    @classmethod
    def _from_arrow(cls, path: Path) -> pd.DataFrame:
        directory = context_file_loading.get().get("directory", Path("."))
        return pd.read_feather(directory / path, dtype_backend="pyarrow")

    def sort(self):
        """Sort the table as required.

        Sorting is done automatically before writing the table.
        """
        if self.df is not None:
            df = self.df.sort_values(self._sort_keys, ignore_index=True)
            df.index.rename("fid", inplace=True)
            self.df = df  # trigger validation and thus index coercion to int32

    @classmethod
    def tableschema(cls) -> TableT:
        """Retrieve Pandera Schema.

        The type of the field `df` is known to always be an DataFrame[TableT]]] | None
        """
        optionalfieldtype = cls.model_fields["df"].annotation
        fieldtype = optionalfieldtype.__args__[0]  # type: ignore
        T: TableT = fieldtype.__args__[0]
        return T

    @classmethod
    def columns(cls) -> list[str]:
        """Retrieve column names."""
        T = cls.tableschema()
        return list(T.to_schema().columns.keys())

    def __repr__(self) -> str:
        # Make sure not to return just "None", because it gets extremely confusing
        # when debugging.
        return f"{self.tablename()}\n{self.df.__repr__()}"

    def _repr_html_(self):
        if self.df is None:
            return self.__repr__()
        else:
            return f"<div>{self.tablename()}</div>" + self.df._repr_html_()

    def __getitem__(self, index) -> pd.DataFrame | gpd.GeoDataFrame:
        tablename = self.tablename()
        if self.df is None:
            raise ValueError(f"Cannot index into {tablename}: it contains no data.")

        # Allow for indexing with multiple values.
        np_index = np.atleast_1d(index)
        missing = np.setdiff1d(np_index, self.df["node_id"].unique())
        if missing.size > 0:
            raise IndexError(f"{tablename} does not contain node_id: {missing}")

        # Index with .loc[..., :] to always return a DataFrame.
        return self.df.loc[self.df["node_id"].isin(np_index), :]


class SpatialTableModel(TableModel[TableT], Generic[TableT]):
    df: GeoDataFrame[TableT] | None = Field(default=None, exclude=True, repr=False)

    def sort(self):
        # Only sort the index (node_id / link_id) since this needs to be sorted in a GeoPackage.
        # Under most circumstances, this retains the input order,
        # making the link_id as stable as possible; useful for post-processing.
        self.df.sort_index(inplace=True)

    @classmethod
    def _from_db(cls, path: Path, table: str):
        with closing(connect(path)) as connection:
            if exists(connection, table):
                # pyogrio hardcodes fid name on reading
                df = gpd.read_file(
                    path,
                    layer=table,
                    engine="pyogrio",
                    fid_as_index=True,
                    use_arrow=True,
                    # tell pyarrow to map to pd.ArrowDtype rather than NumPy
                    arrow_to_pandas_kwargs={"types_mapper": pd.ArrowDtype},
                )
            else:
                df = None

            return df

    def _write_geopackage(self, path: Path) -> None:
        """
        Write the contents of the input to the GeoPackage.

        Parameters
        ----------
        path : Path
        """
        assert self.df is not None
        self.df.to_file(
            path,
            layer=self.tablename(),
            driver="GPKG",
            index=True,
            fid=self.df.index.name,
            engine="pyogrio",
        )
        _add_styles_to_geopackage(path, self.tablename())


class ChildModel(BaseModel):
    _parent: Any | None = None
    _parent_field: str | None = None

    @model_validator(mode="after")
    def _check_parent(self) -> "ChildModel":
        if self._parent is not None:
            self._parent.model_fields_set.update({self._parent_field})
        return self


class NodeModel(ChildModel):
    """Base class to handle combining the tables for a single node type."""

    @model_serializer(mode="wrap")
    def set_modeld(
        self, serializer: Callable[[type["NodeModel"]], dict[str, Any]]
    ) -> dict[str, Any]:
        content = serializer(self)
        return dict(filter(lambda x: x[1], content.items()))

    @field_validator("*")
    @classmethod
    def set_sort_keys(cls, v: Any, info: ValidationInfo) -> Any:
        """Set sort keys for all TableModels if present in FieldInfo."""
        if isinstance(v, TableModel):
            field = cls.model_fields[getattr(info, "field_name")]
            extra = field.json_schema_extra
            if extra is not None and isinstance(extra, dict):
                # We set sort_keys ourselves as list[str] in json_schema_extra
                # but mypy doesn't know.
                v._sort_keys = cast(list[str], extra.get("sort_keys", []))
        return v

    @classmethod
    def get_input_type(cls):
        return cls.__name__

    @classmethod
    def _layername(cls, field: str) -> str:
        return f"{cls.get_input_type()}{delimiter}{field}"

    def _tables(self) -> Generator[TableModel[Any], Any, None]:
        for key in self._fields():
            attr = getattr(self, key)
            if (
                isinstance(attr, TableModel)
                and (attr.df is not None)
                and not (isinstance(attr, ribasim.geometry.node.NodeTable))
            ):
                yield attr

    def _node_ids(self) -> set[int]:
        node_ids: set[int] = set()
        for table in self._tables():
            node_ids.update(table._node_ids())
        return node_ids

    def _save(self, directory: DirectoryPath, input_dir: DirectoryPath):
        for table in self._tables():
            table._save(directory, input_dir)

    def _repr_content(self) -> str:
        """Generate a succinct overview of the content.

        Skip "empty" attributes: when the dataframe of a TableModel is None.
        """
        content = []
        for field in self._fields():
            attr = getattr(self, field)
            if isinstance(attr, TableModel):
                if attr.df is not None:
                    content.append(field)
            else:
                content.append(field)
        return ", ".join(content)

    def __repr__(self) -> str:
        content = self._repr_content()
        typename = type(self).__name__
        return f"{typename}({content})"



import numbers
from collections.abc import Sequence
from enum import Enum
from typing import Any

import numpy as np
import pandas as pd
import pydantic
from geopandas import GeoDataFrame
from pydantic import ConfigDict, Field, NonNegativeInt, model_validator
from shapely.geometry import Point

from ribasim.geometry import BasinAreaSchema, NodeTable
from ribasim.geometry.link import NodeData
from ribasim.input_base import ChildModel, NodeModel, SpatialTableModel, TableModel

# These schemas are autogenerated
from ribasim.schemas import (
    BasinConcentrationExternalSchema,
    BasinConcentrationSchema,
    BasinConcentrationStateSchema,
    BasinProfileSchema,
    BasinStateSchema,
    BasinStaticSchema,
    BasinSubgridSchema,
    BasinSubgridTimeSchema,
    BasinTimeSchema,
    ContinuousControlFunctionSchema,
    ContinuousControlVariableSchema,
    DiscreteControlConditionSchema,
    DiscreteControlLogicSchema,
    DiscreteControlVariableSchema,
    FlowBoundaryConcentrationSchema,
    FlowBoundaryStaticSchema,
    FlowBoundaryTimeSchema,
    FlowDemandStaticSchema,
    FlowDemandTimeSchema,
    LevelBoundaryConcentrationSchema,
    LevelBoundaryStaticSchema,
    LevelBoundaryTimeSchema,
    LevelDemandStaticSchema,
    LevelDemandTimeSchema,
    LinearResistanceStaticSchema,
    ManningResistanceStaticSchema,
    OutletStaticSchema,
    OutletTimeSchema,
    PidControlStaticSchema,
    PidControlTimeSchema,
    PumpStaticSchema,
    PumpTimeSchema,
    TabulatedRatingCurveStaticSchema,
    TabulatedRatingCurveTimeSchema,
    UserDemandConcentrationSchema,
    UserDemandStaticSchema,
    UserDemandTimeSchema,
)
from ribasim.utils import _concat, _pascal_to_snake


class SourcePriority(ChildModel):
    """
    Specify per source node type what its default source priority is.

    flow_boundary and level_boundary nodes are combined into the single category 'boundary'.
    """

    user_demand: int = 1000
    boundary: int = 2000
    level_demand: int = 3000
    flow_demand: int = 4000
    subnetwork_inlet: int = 5000


class Allocation(ChildModel):
    """
    Defines the allocation optimization algorithm options.

    Attributes
    ----------
    timestep : float
        The simulated time in seconds between successive allocation calls (Optional, defaults to 86400)
    use_allocation : bool
        Whether the allocation algorithm should be active. If not, `UserDemand` nodes attempt to
        abstract their full demand (Optional, defaults to False)
    """

    timestep: float = 86400.0
    use_allocation: bool = False
    default_source_priority: SourcePriority = SourcePriority()


class Results(ChildModel):
    compression: bool = True
    compression_level: int = 6
    subgrid: bool = False


class Solver(ChildModel):
    """
    Defines the numerical solver options.

    For more details see <https://docs.sciml.ai/DiffEqDocs/stable/basics/common_solver_opts/#solver_options>.

    Attributes
    ----------
    algorithm : str
        The used numerical time integration algorithm (Optional, defaults to QNDF)
    saveat : float
        Time interval in seconds between saves of output data.
        0 saves every timestep, inf only saves at start- and endtime. (Optional, defaults to 86400)
    dt : float
        Timestep of the solver. (Optional, defaults to None which implies adaptive timestepping)
    dtmin : float
        The minimum allowed timestep of the solver (Optional, defaults to 0.0)
    dtmax : float
        The maximum allowed timestep size (Optional, defaults to 0.0 which implies the total length of the simulation)
    force_dtmin : bool
        If a smaller dt than dtmin is needed to meet the set error tolerances, the simulation stops, unless force_dtmin = true
        (Optional, defaults to False)
    abstol : float
        The absolute tolerance for adaptive timestepping (Optional, defaults to 1e-7)
    reltol : float
        The relative tolerance for adaptive timestepping (Optional, defaults to 1e-7)
    maxiters : int
        The total number of linear iterations over the whole simulation. (Defaults to 1e9, only needs to be increased for extremely long simulations)
    sparse : bool
        Whether a sparse Jacobian matrix is used, which gives a significant speedup for models with >~10 basins.
    autodiff : bool
        Whether automatic differentiation instead of fine difference is used to compute the Jacobian. (Optional, defaults to true)
    """

    algorithm: str = "QNDF"
    saveat: float = 86400.0
    dt: float | None = None
    dtmin: float | None = None
    dtmax: float | None = None
    force_dtmin: bool = False
    abstol: float = 1e-06
    reltol: float = 1e-05
    maxiters: int = 1000000000
    sparse: bool = True
    autodiff: bool = False
    evaporate_mass: bool = True


class Verbosity(str, Enum):
    debug = "debug"
    info = "info"
    warn = "warn"
    error = "error"


class Logging(ChildModel):
    """
    Defines the logging behavior of the core.

    Attributes
    ----------
    verbosity : Verbosity
        The verbosity of the logging: debug/info/warn/error (Optional, defaults to info)
    """

    verbosity: Verbosity = Verbosity.info


class Experimental(ChildModel):
    """
    Defines experimental features.

    Attributes
    ----------
    concentration : bool
        Whether to enable tracer support (default is False)
    """

    concentration: bool = False


class Node(pydantic.BaseModel):
    """
    Defines a node for the model.

    Attributes
    ----------
    node_id : Optional[NonNegativeInt]
        Integer ID of the node. Must be unique for the model.
    geometry : shapely.geometry.Point
        The coordinates of the node.
    name : str
        An optional name of the node.
    subnetwork_id : int
        Optionally adds this node to a subnetwork, which is input for the allocation algorithm.
    source_priority : int
        Optionally overrides the source priority for this node, which is used in the allocation algorithm.
    cyclic_time : bool
        Optionally extrapolate forcing timeseries periodically. Defaults to False.
    """

    node_id: NonNegativeInt | None = None
    geometry: Point
    name: str = ""
    subnetwork_id: int | None = None
    source_priority: int | None = None
    cyclic_time: bool = False

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")

    def __init__(
        self,
        node_id: NonNegativeInt | None = None,
        geometry: Point = Point(),
        **kwargs,
    ) -> None:
        if geometry.is_empty:
            raise (ValueError("Node geometry must be a valid Point"))
        elif geometry.has_z:
            # Remove any Z coordinate, this will cause issues connecting 2D and 3D nodes
            geometry = Point(geometry.x, geometry.y)
        super().__init__(node_id=node_id, geometry=geometry, **kwargs)

    def into_geodataframe(self, node_type: str, node_id: int) -> GeoDataFrame:
        extra = self.model_extra if self.model_extra is not None else {}
        gdf = GeoDataFrame(
            data={
                "node_id": pd.Series([node_id], dtype=np.int32),
                "node_type": pd.Series([node_type], dtype=str),
                "name": pd.Series([self.name], dtype=str),
                "subnetwork_id": pd.Series([self.subnetwork_id], dtype=pd.Int32Dtype()),
                "source_priority": pd.Series(
                    [self.source_priority], dtype=pd.Int32Dtype()
                ),
                "cyclic_time": pd.Series([self.cyclic_time], dtype=bool),
                **extra,
            },
            geometry=[self.geometry],
        )
        gdf.set_index("node_id", inplace=True)
        return gdf


class MultiNodeModel(NodeModel):
    node: NodeTable = Field(default_factory=NodeTable)
    _node_type: str

    @model_validator(mode="after")
    def filter(self) -> "MultiNodeModel":
        self.node.filter(self.__class__.__name__)
        return self

    def add(
        self, node: Node, tables: Sequence[TableModel[Any]] | None = None
    ) -> NodeData:
        """Add a node and the associated data to the model.

        Parameters
        ----------
        node : Ribasim.Node
        tables : Sequence[TableModel[Any]] | None

        Raises
        ------
        ValueError
            When the given node ID already exists
        """
        if tables is None:
            tables = []

        node_id = node.node_id

        if self._parent is None:
            raise ValueError(
                f"You can only add to a {self._node_type} MultiNodeModel when attached to a Model."
            )

        if node_id is None:
            node_id = self._parent._used_node_ids.new_id()
        elif node_id in self._parent._used_node_ids:
            raise ValueError(
                f"Node IDs have to be unique, but {node_id} already exists."
            )

        for table in tables:
            member_name = _pascal_to_snake(table.__class__.__name__)
            existing_member = getattr(self, member_name)
            existing_table = (
                existing_member.df if existing_member.df is not None else pd.DataFrame()
            )
            assert table.df is not None
            table_to_append = table.df.assign(node_id=node_id)
            if isinstance(table_to_append, GeoDataFrame):
                table_to_append.set_crs(self._parent.crs, inplace=True)
            new_table = _concat([existing_table, table_to_append], ignore_index=True)
            setattr(self, member_name, new_table)

        node_table = node.into_geodataframe(
            node_type=self.__class__.__name__, node_id=node_id
        )
        node_table.set_crs(self._parent.crs, inplace=True)
        if self.node.df is None:
            self.node.df = node_table
        else:
            df = _concat([self.node.df, node_table])
            self.node.df = df

        self._parent._used_node_ids.add(node_id)
        return self[node_id]

    def __getitem__(self, index: int) -> NodeData:
        # Unlike TableModel, support only indexing single rows.
        if not isinstance(index, numbers.Integral):
            node_model_name = type(self).__name__
            indextype = type(index).__name__
            raise TypeError(
                f"{node_model_name} index must be an integer, not {indextype}"
            )

        row = self.node.df.loc[index]
        return NodeData(
            node_id=int(index), node_type=row["node_type"], geometry=row["geometry"]
        )


class Terminal(MultiNodeModel): ...


class PidControl(MultiNodeModel):
    static: TableModel[PidControlStaticSchema] = Field(
        default_factory=TableModel[PidControlStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state"]},
    )
    time: TableModel[PidControlTimeSchema] = Field(
        default_factory=TableModel[PidControlTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "time"]},
    )


class LevelBoundary(MultiNodeModel):
    static: TableModel[LevelBoundaryStaticSchema] = Field(
        default_factory=TableModel[LevelBoundaryStaticSchema],
        json_schema_extra={"sort_keys": ["node_id"]},
    )
    time: TableModel[LevelBoundaryTimeSchema] = Field(
        default_factory=TableModel[LevelBoundaryTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "time"]},
    )
    concentration: TableModel[LevelBoundaryConcentrationSchema] = Field(
        default_factory=TableModel[LevelBoundaryConcentrationSchema],
        json_schema_extra={"sort_keys": ["node_id", "substance", "time"]},
    )


class Pump(MultiNodeModel):
    static: TableModel[PumpStaticSchema] = Field(
        default_factory=TableModel[PumpStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state"]},
    )
    time: TableModel[PumpTimeSchema] = Field(
        default_factory=TableModel[PumpTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "time"]},
    )


class TabulatedRatingCurve(MultiNodeModel):
    static: TableModel[TabulatedRatingCurveStaticSchema] = Field(
        default_factory=TableModel[TabulatedRatingCurveStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state", "level"]},
    )
    time: TableModel[TabulatedRatingCurveTimeSchema] = Field(
        default_factory=TableModel[TabulatedRatingCurveTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "time", "level"]},
    )


class UserDemand(MultiNodeModel):
    static: TableModel[UserDemandStaticSchema] = Field(
        default_factory=TableModel[UserDemandStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "demand_priority"]},
    )
    time: TableModel[UserDemandTimeSchema] = Field(
        default_factory=TableModel[UserDemandTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "demand_priority", "time"]},
    )
    concentration: TableModel[UserDemandConcentrationSchema] = Field(
        default_factory=TableModel[UserDemandConcentrationSchema],
        json_schema_extra={"sort_keys": ["node_id", "substance", "time"]},
    )


class LevelDemand(MultiNodeModel):
    static: TableModel[LevelDemandStaticSchema] = Field(
        default_factory=TableModel[LevelDemandStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "demand_priority"]},
    )
    time: TableModel[LevelDemandTimeSchema] = Field(
        default_factory=TableModel[LevelDemandTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "demand_priority", "time"]},
    )


class FlowBoundary(MultiNodeModel):
    static: TableModel[FlowBoundaryStaticSchema] = Field(
        default_factory=TableModel[FlowBoundaryStaticSchema],
        json_schema_extra={"sort_keys": ["node_id"]},
    )
    time: TableModel[FlowBoundaryTimeSchema] = Field(
        default_factory=TableModel[FlowBoundaryTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "time"]},
    )
    concentration: TableModel[FlowBoundaryConcentrationSchema] = Field(
        default_factory=TableModel[FlowBoundaryConcentrationSchema],
        json_schema_extra={"sort_keys": ["node_id", "substance", "time"]},
    )


class FlowDemand(MultiNodeModel):
    static: TableModel[FlowDemandStaticSchema] = Field(
        default_factory=TableModel[FlowDemandStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "demand_priority"]},
    )
    time: TableModel[FlowDemandTimeSchema] = Field(
        default_factory=TableModel[FlowDemandTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "demand_priority", "time"]},
    )


class Basin(MultiNodeModel):
    profile: TableModel[BasinProfileSchema] = Field(
        default_factory=TableModel[BasinProfileSchema],
        json_schema_extra={"sort_keys": ["node_id", "level"]},
    )
    state: TableModel[BasinStateSchema] = Field(
        default_factory=TableModel[BasinStateSchema],
        json_schema_extra={"sort_keys": ["node_id"]},
    )
    static: TableModel[BasinStaticSchema] = Field(
        default_factory=TableModel[BasinStaticSchema],
        json_schema_extra={"sort_keys": ["node_id"]},
    )
    time: TableModel[BasinTimeSchema] = Field(
        default_factory=TableModel[BasinTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "time"]},
    )
    subgrid: TableModel[BasinSubgridSchema] = Field(
        default_factory=TableModel[BasinSubgridSchema],
        json_schema_extra={"sort_keys": ["subgrid_id", "basin_level"]},
    )
    subgrid_time: TableModel[BasinSubgridTimeSchema] = Field(
        default_factory=TableModel[BasinSubgridTimeSchema],
        json_schema_extra={"sort_keys": ["subgrid_id", "time", "basin_level"]},
    )
    area: SpatialTableModel[BasinAreaSchema] = Field(
        default_factory=SpatialTableModel[BasinAreaSchema],
        json_schema_extra={"sort_keys": ["node_id"]},
    )
    concentration: TableModel[BasinConcentrationSchema] = Field(
        default_factory=TableModel[BasinConcentrationSchema],
        json_schema_extra={"sort_keys": ["node_id", "substance", "time"]},
    )
    concentration_external: TableModel[BasinConcentrationExternalSchema] = Field(
        default_factory=TableModel[BasinConcentrationExternalSchema],
        json_schema_extra={"sort_keys": ["node_id", "substance", "time"]},
    )
    concentration_state: TableModel[BasinConcentrationStateSchema] = Field(
        default_factory=TableModel[BasinConcentrationStateSchema],
        json_schema_extra={"sort_keys": ["node_id", "substance"]},
    )


class ManningResistance(MultiNodeModel):
    static: TableModel[ManningResistanceStaticSchema] = Field(
        default_factory=TableModel[ManningResistanceStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state"]},
    )


class DiscreteControl(MultiNodeModel):
    variable: TableModel[DiscreteControlVariableSchema] = Field(
        default_factory=TableModel[DiscreteControlVariableSchema],
        json_schema_extra={
            "sort_keys": [
                "node_id",
                "compound_variable_id",
                "listen_node_id",
                "variable",
            ]
        },
    )
    condition: TableModel[DiscreteControlConditionSchema] = Field(
        default_factory=TableModel[DiscreteControlConditionSchema],
        json_schema_extra={
            "sort_keys": ["node_id", "compound_variable_id", "condition_id"]
        },
    )
    logic: TableModel[DiscreteControlLogicSchema] = Field(
        default_factory=TableModel[DiscreteControlLogicSchema],
        json_schema_extra={"sort_keys": ["node_id", "truth_state"]},
    )


class Outlet(MultiNodeModel):
    static: TableModel[OutletStaticSchema] = Field(
        default_factory=TableModel[OutletStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state"]},
    )
    time: TableModel[OutletTimeSchema] = Field(
        default_factory=TableModel[OutletTimeSchema],
        json_schema_extra={"sort_keys": ["node_id", "time"]},
    )


class LinearResistance(MultiNodeModel):
    static: TableModel[LinearResistanceStaticSchema] = Field(
        default_factory=TableModel[LinearResistanceStaticSchema],
        json_schema_extra={"sort_keys": ["node_id", "control_state"]},
    )


class ContinuousControl(MultiNodeModel):
    variable: TableModel[ContinuousControlVariableSchema] = Field(
        default_factory=TableModel[ContinuousControlVariableSchema],
        json_schema_extra={"sort_keys": ["node_id", "listen_node_id", "variable"]},
    )
    function: TableModel[ContinuousControlFunctionSchema] = Field(
        default_factory=TableModel[ContinuousControlFunctionSchema],
        json_schema_extra={"sort_keys": ["node_id", "input"]},
    )



from pathlib import Path
from typing import NamedTuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pandera as pa
import shapely
from matplotlib.axes import Axes
from numpy.typing import NDArray
from pandera.dtypes import Int32
from pandera.typing import Index, Series
from pandera.typing.geopandas import GeoDataFrame, GeoSeries
from pydantic import NonNegativeInt, PrivateAttr, model_validator
from shapely.geometry import LineString, MultiLineString, Point

from ribasim.db_utils import _get_db_schema_version
from ribasim.input_base import SpatialTableModel
from ribasim.utils import UsedIDs, _concat
from ribasim.validation import (
    can_connect,
    control_link_neighbor_amount,
    flow_link_neighbor_amount,
    node_type_connectivity,
)

from .base import _GeoBaseSchema

__all__ = ("LinkTable",)

SPATIALCONTROLNODETYPES = {
    "ContinuousControl",
    "DiscreteControl",
    "FlowDemand",
    "LevelDemand",
    "PidControl",
}


class NodeData(NamedTuple):
    node_id: int
    node_type: str
    geometry: Point

    def __repr__(self) -> str:
        return f"{self.node_type} #{self.node_id}"


class LinkSchema(_GeoBaseSchema):
    link_id: Index[Int32] = pa.Field(default=0, ge=0, check_name=True)
    name: Series[str] = pa.Field(default="")
    from_node_id: Series[Int32] = pa.Field(default=0)
    to_node_id: Series[Int32] = pa.Field(default=0)
    link_type: Series[str] = pa.Field(default="flow")
    geometry: GeoSeries[LineString] = pa.Field(default=None, nullable=True)

    @classmethod
    def _index_name(self) -> str:
        return "link_id"


class LinkTable(SpatialTableModel[LinkSchema]):
    """Defines the connections between nodes."""

    _used_link_ids: UsedIDs = PrivateAttr(default_factory=UsedIDs)

    @model_validator(mode="after")
    def _update_used_ids(self) -> "LinkTable":
        if self.df is not None and len(self.df.index) > 0:
            self._used_link_ids.node_ids.update(self.df.index)
            self._used_link_ids.max_node_id = self.df.index.max()
        return self

    @classmethod
    def _from_db(cls, path: Path, table: str) -> pd.DataFrame | None:
        schema_version = _get_db_schema_version(path)
        # The table name was changed from "Edge" to "Link" in schema_version 4.
        if schema_version < 4:
            table = "Edge"
        return super()._from_db(path, table)

    def add(
        self,
        from_node: NodeData,
        to_node: NodeData,
        geometry: LineString | MultiLineString | None = None,
        name: str = "",
        link_id: NonNegativeInt | None = None,
        **kwargs,
    ):
        """
        Add an link between nodes.

        The type of the link (flow or control) is automatically inferred from the type of the `from_node`.

        Parameters
        ----------
        from_node : NodeData
            A node indexed by its node ID, e.g. `model.basin[1]`
        to_node: NodeData
            A node indexed by its node ID, e.g. `model.linear_resistance[1]`
        geometry : LineString | MultiLineString | None
            The geometry of a line. If not supplied, it creates a straight line between the nodes.
        name : str
            An optional name for the link.
        link_id : int
            An optional non-negative link ID. If not supplied, it will be automatically generated.
        **kwargs : Dict
        """
        if not can_connect(from_node.node_type, to_node.node_type):
            raise ValueError(
                f"Node #{to_node.node_id} of type {to_node.node_type} cannot be downstream of node #{from_node.node_id} of type {from_node.node_type}. Possible downstream node types: {node_type_connectivity[from_node.node_type]}."
            )

        if self.df is not None:
            if (
                "UserDemand" not in [from_node.node_type, to_node.node_type]
                and not self.df[
                    (self.df.from_node_id == to_node.node_id)
                    & (self.df.to_node_id == from_node.node_id)
                ].empty
            ):
                raise ValueError(
                    f"Link ({link_id=}, {from_node=}, {to_node=}) is not allowed since the opposite link already exists (this is only allowed for UserDemand)."
                )

        geometry_to_append = (
            [LineString([from_node.geometry, to_node.geometry])]
            if geometry is None
            else [geometry]
        )
        link_type = (
            "control" if from_node.node_type in SPATIALCONTROLNODETYPES else "flow"
        )
        self._validate_link(to_node, from_node, link_type)
        assert self.df is not None
        if link_id is None:
            link_id = self._used_link_ids.new_id()
        elif link_id in self._used_link_ids:
            raise ValueError(
                f"Link IDs have to be unique, but {link_id} already exists."
            )

        table_to_append = GeoDataFrame[LinkSchema](
            data={
                "from_node_id": [from_node.node_id],
                "to_node_id": [to_node.node_id],
                "link_type": [link_type],
                "name": [name],
                **kwargs,
            },
            geometry=geometry_to_append,
            crs=self.df.crs,
            index=pd.Index([link_id], name="link_id"),
        )

        self.df = GeoDataFrame[LinkSchema](_concat([self.df, table_to_append]))
        if self.df.duplicated(subset=["from_node_id", "to_node_id"]).any():
            raise ValueError(
                f"Links have to be unique, but link with from_node_id {from_node.node_id} to_node_id {to_node.node_id} already exists."
            )
        self._used_link_ids.add(link_id)

    def _validate_link(self, to_node: NodeData, from_node: NodeData, link_type: str):
        assert self.df is not None
        in_neighbor: int = self.df.loc[
            (self.df["to_node_id"] == to_node.node_id)
            & (self.df["link_type"] == link_type)
        ].shape[0]

        out_neighbor: int = self.df.loc[
            (self.df["from_node_id"] == from_node.node_id)
            & (self.df["link_type"] == link_type)
        ].shape[0]
        # validation on neighbor amount
        max_in_flow: int = flow_link_neighbor_amount[to_node.node_type][1]
        max_out_flow: int = flow_link_neighbor_amount[from_node.node_type][3]
        max_in_control: int = control_link_neighbor_amount[to_node.node_type][1]
        max_out_control: int = control_link_neighbor_amount[from_node.node_type][3]
        if link_type == "flow":
            if in_neighbor >= max_in_flow:
                raise ValueError(
                    f"Node {to_node.node_id} can have at most {max_in_flow} flow link inneighbor(s) (got {in_neighbor})"
                )
            if out_neighbor >= max_out_flow:
                raise ValueError(
                    f"Node {from_node.node_id} can have at most {max_out_flow} flow link outneighbor(s) (got {out_neighbor})"
                )
        elif link_type == "control":
            if in_neighbor >= max_in_control:
                raise ValueError(
                    f"Node {to_node.node_id} can have at most {max_in_control} control link inneighbor(s) (got {in_neighbor})"
                )
            if out_neighbor >= max_out_control:
                raise ValueError(
                    f"Node {from_node.node_id} can have at most {max_out_control} control link outneighbor(s) (got {out_neighbor})"
                )

    def _get_where_link_type(self, link_type: str) -> NDArray[np.bool_]:
        assert self.df is not None
        return (self.df.link_type == link_type).to_numpy()

    def plot(self, **kwargs) -> Axes:
        """Plot the links of the model.

        Parameters
        ----------
        **kwargs : Dict
            Supported: 'ax', 'color_flow', 'color_control'
        """
        assert self.df is not None
        kwargs = kwargs.copy()  # Avoid side-effects
        ax = kwargs.get("ax", None)
        color_flow = kwargs.pop("color_flow", None)
        color_control = kwargs.pop("color_control", None)

        if ax is None:
            _, ax = plt.subplots()
            ax.axis("off")
            kwargs["ax"] = ax

        kwargs_flow = kwargs.copy()
        kwargs_control = kwargs.copy()

        if color_flow is None:
            color_flow = "#3690c0"  # lightblue
            kwargs_flow["color"] = color_flow
            kwargs_flow["label"] = "Flow link"
        if color_control is None:
            color_control = "grey"
            kwargs_control["color"] = color_control
            kwargs_control["label"] = "Control link"

        where_flow = self._get_where_link_type("flow")
        where_control = self._get_where_link_type("control")

        if not self.df[where_flow].empty:
            self.df[where_flow].plot(**kwargs_flow)

        if where_control.any():
            self.df[where_control].plot(**kwargs_control)

        # Determine the angle for every caret marker and where to place it.
        coords, index = shapely.get_coordinates(self.df.geometry, return_index=True)
        keep = np.diff(index) == 0
        link_coords = np.stack((coords[:-1, :], coords[1:, :]), axis=1)[keep]
        x, y = np.mean(link_coords, axis=1).T
        dx, dy = np.diff(link_coords, axis=1)[:, 0, :].T
        angle = np.degrees(np.arctan2(dy, dx)) - 90

        # Set the color of the marker to match the line.
        # Black is default, set color_flow otherwise; then set color_control.
        color_index = index[1:][keep]
        color = np.where(where_flow[color_index], color_flow, "k")
        color = np.where(where_control[color_index], color_control, color)

        # A faster alternative may be ax.quiver(). However, getting the scaling
        # right is tedious.
        for m_x, m_y, m_angle, c in zip(x, y, angle, color):
            ax.plot(
                m_x,
                m_y,
                marker=(3, 0, m_angle),
                markersize=5,
                linestyle="None",
                c=c,
            )

        return ax

    def __getitem__(self, _):
        raise NotImplementedError



from ribasim.input_base import TableModel
from ribasim.schemas import (
    FlowDemandStaticSchema,
    FlowDemandTimeSchema,
)

__all__ = ["Static", "Time"]


class Static(TableModel[FlowDemandStaticSchema]):
    pass


class Time(TableModel[FlowDemandTimeSchema]):
    pass



from ribasim.input_base import TableModel
from ribasim.schemas import PumpStaticSchema, PumpTimeSchema

__all__ = ["Static", "Time"]


class Static(TableModel[PumpStaticSchema]):
    pass


class Time(TableModel[PumpTimeSchema]):
    pass



from ribasim.input_base import TableModel
from ribasim.schemas import (
    DiscreteControlConditionSchema,
    DiscreteControlLogicSchema,
    DiscreteControlVariableSchema,
)

__all__ = ["Condition", "Logic", "Variable"]


class Variable(TableModel[DiscreteControlVariableSchema]):
    pass


class Condition(TableModel[DiscreteControlConditionSchema]):
    pass


class Logic(TableModel[DiscreteControlLogicSchema]):
    pass



from ribasim.nodes import (
    basin,
    continuous_control,
    discrete_control,
    flow_boundary,
    level_boundary,
    level_demand,
    linear_resistance,
    manning_resistance,
    outlet,
    pid_control,
    pump,
    tabulated_rating_curve,
    user_demand,
)

__all__ = [
    "basin",
    "continuous_control",
    "discrete_control",
    "flow_boundary",
    "level_boundary",
    "level_demand",
    "linear_resistance",
    "manning_resistance",
    "outlet",
    "pid_control",
    "pump",
    "tabulated_rating_curve",
    "user_demand",
]



from ribasim.input_base import TableModel
from ribasim.schemas import (
    LevelDemandStaticSchema,
    LevelDemandTimeSchema,
)

__all__ = ["Static", "Time"]


class Static(TableModel[LevelDemandStaticSchema]):
    pass


class Time(TableModel[LevelDemandTimeSchema]):
    pass



from pathlib import Path

import pytest
from ribasim import Model
from ribasim.db_utils import _get_db_schema_version

root_folder = Path(__file__).parent.parent.parent.parent
print(root_folder)


@pytest.mark.regression
def test_hws_migration(tmp_path):
    toml_path = root_folder / "models/hws_migration_test/hws.toml"
    db_path = root_folder / "models/hws_migration_test/database.gpkg"

    assert (
        toml_path.exists()
    ), "Can't find the model, did you retrieve it with get_benchmark.py?"

    assert _get_db_schema_version(db_path) == 0
    model = Model.read(toml_path)

    assert model.link.df.index.name == "link_id"
    assert len(model.link.df) == 454
    model.write(tmp_path / "hws_migrated.toml")



import pandera as pa
from pandera.dtypes import Int32
from pandera.typing import Index, Series
from pandera.typing.geopandas import GeoSeries
from shapely.geometry import MultiPolygon, Polygon

from .base import _GeoBaseSchema


class BasinAreaSchema(_GeoBaseSchema):
    fid: Index[Int32] = pa.Field(default=0, check_name=True)
    node_id: Series[Int32] = pa.Field(nullable=False, default=0)
    geometry: GeoSeries[MultiPolygon] = pa.Field(default=None, nullable=True)

    @pa.parser("geometry")
    def convert_to_multi(cls, series):
        return series.apply(
            lambda geom: MultiPolygon([geom]) if isinstance(geom, Polygon) else geom
        )



from ribasim import geometry, nodes
from ribasim.input_base import TableModel
from ribasim.schemas import BasinSubgridSchema


def test_tablemodel_schema():
    schema = TableModel[BasinSubgridSchema].tableschema()
    assert schema == BasinSubgridSchema


def test_tablename():
    cls = nodes.tabulated_rating_curve.Static
    assert cls.tablename() == "TabulatedRatingCurve / static"

    cls = nodes.basin.ConcentrationExternal
    assert cls.tablename() == "Basin / concentration_external"

    cls = geometry.link.LinkTable
    assert cls.tablename() == "Link"



from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import ribasim
import tomli
from pandas import DataFrame
from pandas.testing import assert_frame_equal
from pydantic import ValidationError
from ribasim import Model, Node, Solver
from ribasim.nodes import basin, flow_boundary, flow_demand, pump, user_demand
from ribasim.utils import UsedIDs
from shapely.geometry import Point


def __assert_equal(a: DataFrame, b: DataFrame) -> None:
    """Lenient version of pandas.testing.assert_frame_equal."""
    if a is None and b is None:
        return
    elif a is None or b is None:
        assert False

    a = a.reset_index(drop=True)
    b = b.reset_index(drop=True)

    assert_frame_equal(a, b)


def test_basic(basic, tmp_path):
    model_orig = basic
    toml_path = tmp_path / "basic/ribasim.toml"
    assert model_orig.filepath is None
    model_orig.write(toml_path)
    assert model_orig.filepath == toml_path
    model_loaded = Model.read(toml_path)
    assert model_loaded.filepath == toml_path

    with open(toml_path, "rb") as f:
        toml_dict = tomli.load(f)

    assert toml_dict["ribasim_version"] == ribasim.__version__

    __assert_equal(model_orig.link.df, model_loaded.link.df)
    __assert_equal(model_orig.node_table().df, model_loaded.node_table().df)
    assert model_loaded.basin.time.df is None


def test_basic_arrow(basic_arrow, tmp_path):
    model_orig = basic_arrow
    model_orig.write(tmp_path / "basic_arrow/ribasim.toml")
    model_loaded = Model.read(tmp_path / "basic_arrow/ribasim.toml")

    __assert_equal(model_orig.basin.profile.df, model_loaded.basin.profile.df)


def test_basic_transient(basic_transient, tmp_path):
    model_orig = basic_transient
    model_orig.write(tmp_path / "basic_transient/ribasim.toml")
    model_loaded = Model.read(tmp_path / "basic_transient/ribasim.toml")

    __assert_equal(model_orig.link.df, model_loaded.link.df)

    time = model_loaded.basin.time
    assert model_orig.basin.time.df.time.iloc[0] == time.df.time.iloc[0]
    assert time.df.node_id.dtype == "int32[pyarrow]"
    __assert_equal(model_orig.basin.time.df, time.df)
    assert time.df.shape == (1468, 6)


@pytest.mark.xfail(reason="Needs implementation")
def test_pydantic():
    pass
    # static_data_bad = pd.DataFrame(data={"node_id": [1, 2, 3]})
    # test that it throws on missing flow_rate


def test_repr():
    pump_static = pump.Static(flow_rate=[1.0, -1.0, 0.0])

    assert repr(pump_static).startswith("Pump / static")
    # Ensure _repr_html doesn't error
    assert isinstance(pump_static._repr_html_(), str)


def test_extra_columns():
    pump_static = pump.Static(meta_id=[-1], flow_rate=[1.2])
    assert "meta_id" in pump_static.df.columns
    assert pump_static.df.meta_id.iloc[0] == -1

    with pytest.raises(ValidationError):
        # Extra column "extra" needs "meta_" prefix
        pump.Static(extra=[-2], flow_rate=[1.2])


def test_index_tables():
    p = pump.Static(flow_rate=[1.2])
    assert p.df.index.name == "fid"
    # Index name is applied by _name_index
    df = p.df.reset_index(drop=True)
    assert df.index.name is None
    p.df = df
    assert p.df.index.name == "fid"


def test_extra_spatial_columns():
    model = Model(
        starttime="2020-01-01",
        endtime="2021-01-01",
        crs="EPSG:28992",
        solver=Solver(algorithm="Tsit5"),
    )

    model.basin.add(
        Node(1, Point(0, 0), meta_id=1),
        [basin.Profile(area=1000.0, level=[0.0, 1.0]), basin.State(level=[1.0])],
    )
    node = Node(2, Point(1, 0.5, 3.0), meta_id=2)
    assert not node.geometry.has_z
    model.user_demand.add(
        node,
        [
            user_demand.Static(
                demand=[1e-4], return_factor=0.9, min_level=0.9, demand_priority=1
            )
        ],
    )

    model.link.add(model.basin[1], model.user_demand[2], meta_foo=1)
    assert "meta_foo" in model.link.df.columns
    assert "meta_id" in model.node_table().df.columns

    with pytest.raises(ValidationError):
        model.basin.add(
            Node(3, Point(0, 0), foo=1),
            [basin.Profile(area=1000.0, level=[0.0, 1.0]), basin.State(level=[1.0])],
        )
    with pytest.raises(ValidationError):
        model.user_demand.add(
            Node(4, Point(1, -0.5), meta_id=3),
            [
                user_demand.Static(
                    demand=[1e-4], return_factor=0.9, min_level=0.9, demand_priority=1
                )
            ],
        )
        model.link.add(model.basin[1], model.user_demand[4], foo=1)


def test_link_autoincrement(basic):
    model = basic
    model.link.df = model.link.df.iloc[0:0]  # clear the table
    model.link._used_link_ids = UsedIDs()  # and reset the counter

    model.link.add(model.basin[1], model.manning_resistance[2], link_id=20)
    assert model.link.df.index[-1] == 20

    model.link.add(model.manning_resistance[2], model.basin[3])
    assert model.link.df.index[-1] == 21

    # Can use any remaining positive integer
    model.link.add(model.basin[3], model.tabulated_rating_curve[8], link_id=1)
    assert model.link.df.index[-1] == 1

    with pytest.raises(
        ValueError,
        match="Link IDs have to be unique, but 1 already exists.",
    ):
        model.link.add(
            model.linear_resistance[10],
            model.level_boundary[17],
            link_id=1,
        )


def test_node_autoincrement():
    model = Model(
        starttime="2020-01-01",
        endtime="2021-01-01",
        crs="EPSG:28992",
        solver=Solver(algorithm="Tsit5"),
    )

    model.basin.add(Node(20, Point(0, 0)), [basin.State(level=[1.0])])
    with pytest.raises(
        ValueError, match="Node IDs have to be unique, but 20 already exists."
    ):
        model.user_demand.add(
            Node(20, Point(1, 0.5)),
            [
                user_demand.Static(
                    demand=[1e-4], return_factor=0.9, min_level=0.9, demand_priority=1
                )
            ],
        )

    nbasin = model.basin.add(Node(geometry=Point(0, 0)), [basin.State(level=[1.0])])
    assert nbasin.node_id == 21

    # Can use any remaining positive integer
    model.basin.add(Node(1, geometry=Point(0, 0)), [basin.State(level=[1.0])])
    nbasin = model.basin.add(Node(geometry=Point(0, 0)), [basin.State(level=[1.0])])
    assert nbasin.node_id == 22

    model.basin.add(Node(100, geometry=Point(0, 0)), [basin.State(level=[1.0])])

    nbasin = model.basin.add(Node(geometry=Point(0, 0)), [basin.State(level=[1.0])])
    assert nbasin.node_id == 101


def test_node_autoincrement_existing_model(basic, tmp_path):
    model = basic

    model.write(tmp_path / "ribasim.toml")
    nmodel = Model.read(tmp_path / "ribasim.toml")

    assert nmodel._used_node_ids.max_node_id == 17
    assert nmodel._used_node_ids.node_ids == set(range(1, 18)) - {13}

    assert nmodel.link._used_link_ids.max_node_id == 16
    assert nmodel.link._used_link_ids.node_ids == set(range(1, 17))


def test_node_empty_geometry():
    model = Model(
        starttime="2020-01-01",
        endtime="2021-01-01",
        crs="EPSG:28992",
    )

    with pytest.raises(ValueError, match="Node geometry must be a valid Point"):
        model.user_demand.add(
            Node(),
            [
                user_demand.Static(
                    demand=[1e-4], return_factor=0.9, min_level=0.9, demand_priority=1
                )
            ],
        )
    with pytest.raises(ValueError, match="Node geometry must be a valid Point"):
        model.user_demand.add(
            Node(2),
            [
                user_demand.Static(
                    demand=[1e-4], return_factor=0.9, min_level=0.9, demand_priority=1
                )
            ],
        )


def test_sort(level_range, tmp_path):
    model = level_range
    table = model.discrete_control.condition
    link = model.link

    # apply a wrong sort, then call the sort method to restore order
    table.df.sort_values("greater_than", ascending=False, inplace=True)
    assert table.df.iloc[0]["greater_than"] == 15.0
    assert table._sort_keys == [
        "node_id",
        "compound_variable_id",
        "condition_id",
    ]
    table.sort()
    assert table.df.iloc[0]["greater_than"] == 5.0

    # The link table is not sorted
    assert link.df.iloc[1]["from_node_id"] == 3

    # re-apply wrong sort, then check if it gets sorted on write
    table.df.sort_values("greater_than", ascending=False, inplace=True)
    model.write(tmp_path / "basic/ribasim.toml")
    # write sorts the model in place
    assert table.df.iloc[0]["greater_than"] == 5.0
    model_loaded = ribasim.Model.read(filepath=tmp_path / "basic/ribasim.toml")
    table_loaded = model_loaded.discrete_control.condition
    link_loaded = model_loaded.link
    assert table_loaded.df.iloc[0]["greater_than"] == 5.0
    assert link.df.iloc[1]["from_node_id"] == 3
    __assert_equal(table.df, table_loaded.df)
    __assert_equal(link.df, link_loaded.df)


def test_roundtrip(trivial, tmp_path):
    model1 = trivial
    # set custom Link index
    model1.link.df.index = pd.Index([15, 12], name="link_id")
    model1dir = tmp_path / "model1"
    model2dir = tmp_path / "model2"
    # read a model and then write it to a different path
    model1.write(model1dir / "ribasim.toml")
    model2 = Model.read(model1dir / "ribasim.toml")
    model2.write(model2dir / "ribasim.toml")

    assert (model1dir / "database.gpkg").is_file()
    assert (model2dir / "database.gpkg").is_file()

    assert (model1dir / "ribasim.toml").read_text() == (
        model2dir / "ribasim.toml"
    ).read_text()

    # check if custom Link indexes are retained (sorted)
    assert (model1.link.df.index == [12, 15]).all()
    assert (model2.link.df.index == [12, 15]).all()

    # check if all tables are the same
    __assert_equal(model1.node_table().df, model2.node_table().df)
    __assert_equal(model1.link.df, model2.link.df)
    for node1, node2 in zip(model1._nodes(), model2._nodes()):
        for table1, table2 in zip(node1._tables(), node2._tables()):
            __assert_equal(table1.df, table2.df)


def test_datetime_timezone():
    # Due to a pydantic issue, a time zone was added.
    # https://github.com/Deltares/Ribasim/issues/1282
    model = ribasim.Model(
        starttime="2000-01-01", endtime="2001-01-01 00:00:00", crs="EPSG:28992"
    )
    assert isinstance(model.starttime, datetime)
    assert isinstance(model.endtime, datetime)
    assert model.starttime.tzinfo is None
    assert model.endtime.tzinfo is None


def test_minimal_toml():
    # Check if the TOML used in QGIS tests is still valid.
    toml_path = Path(__file__).parents[3] / "ribasim_qgis/tests/data/simple_valid.toml"
    (toml_path.parent / "database.gpkg").touch()  # database file must exist for `read`
    model = ribasim.Model.read(toml_path)
    assert model.crs == "EPSG:28992"


def test_closed_model(basic, tmp_path):
    # Test whether we can write to a just opened model
    # implicitly testing that the database is closed after read
    toml_path = tmp_path / "basic/ribasim.toml"
    basic.write(toml_path)
    model = ribasim.Model.read(toml_path)
    model.write(toml_path)


def test_arrow_dtype():
    # Below millisecond precision is not supported
    with pytest.raises(ValidationError):
        flow_boundary.Time(
            time=["2021-01-01 00:00:00.1234"],
            flow_rate=np.ones(1),
        )

    # Extra columns don't get coerced to Arrow types
    df = flow_boundary.Time(
        time=["2021-01-01 00:00:00.123", "2021-01-01 00:00:00.456"],
        flow_rate=[1, 2.2],
        meta_obj=["foo", "bar"],
        meta_str=pd.Series(["a", pd.NA], dtype="string[pyarrow]"),
    ).df

    assert (df["node_id"] == 0).all()
    assert df["node_id"].dtype == "int32[pyarrow]"
    assert df["time"].dtype == "timestamp[ms][pyarrow]"
    assert df["time"].dt.tz is None
    assert df["time"].diff().iloc[1] == pd.Timedelta("333ms")
    assert df["flow_rate"].dtype == "double[pyarrow]"
    assert df["meta_obj"].dtype == object
    assert df["meta_str"].dtype == "string[pyarrow]"
    assert df["meta_str"].isna().iloc[1]

    # Check a string column that is part of the schema and a boolean column
    df = pump.Static(
        flow_rate=np.ones(2),
        control_state=["foo", pd.NA],
        active=[None, False],
    ).df

    assert df["control_state"].dtype == "string[pyarrow]"
    assert df["active"].dtype == "bool[pyarrow]"
    assert df["active"].isna().iloc[0]

    # Optional integer column
    df = flow_demand.Static(
        demand=[1, 2.2],
        demand_priority=[1, pd.NA],
    ).df

    assert df["demand_priority"].dtype == "int32[pyarrow]"
    assert df["demand_priority"].isna().iloc[1]

    # Missing optional integer column
    df = flow_demand.Static(
        demand=[1, 2.2],
    ).df

    assert df["demand_priority"].dtype == "int32[pyarrow]"
    assert df["demand_priority"].isna().all()



import re

import pytest
from ribasim import Node
from ribasim.config import Solver
from ribasim.model import Model
from ribasim.nodes import (
    basin,
    level_boundary,
    outlet,
    pid_control,
    pump,
)
from shapely import MultiPolygon, Point, Polygon


def test_duplicate_link(basic):
    model = basic
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Links have to be unique, but link with from_node_id 16 to_node_id 1 already exists."
        ),
    ):
        model.link.add(
            model.flow_boundary[16],
            model.basin[1],
            name="duplicate",
        )


def test_connectivity(trivial):
    model = trivial
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Node #2147483647 of type Terminal cannot be downstream of node #6 of type Basin. Possible downstream node types: ['LinearResistance', 'UserDemand', 'Outlet', 'TabulatedRatingCurve', 'ManningResistance', 'Pump']"
        ),
    ):
        model.link.add(model.basin[6], model.terminal[2147483647])


def test_maximum_flow_neighbor(outlet):
    model = outlet
    with pytest.raises(
        ValueError,
        match=re.escape("Node 2 can have at most 1 flow link outneighbor(s) (got 1)"),
    ):
        model.basin.add(
            Node(4, Point(1.0, 1.0)),
            [
                basin.Profile(area=[1000.0, 1000.0], level=[0.0, 1.0]),
                basin.State(level=[0.0]),
            ],
        )
        model.link.add(model.outlet[2], model.basin[4])

    with pytest.raises(
        ValueError,
        match=re.escape("Node 2 can have at most 1 flow link inneighbor(s) (got 1)"),
    ):
        model.level_boundary.add(
            Node(5, Point(0.0, 1.0)),
            [level_boundary.Static(level=[3.0])],
        )
        model.link.add(model.level_boundary[5], model.outlet[2])


def test_maximum_control_neighbor(pid_control_equation):
    model = pid_control_equation
    with pytest.raises(
        ValueError,
        match=re.escape("Node 2 can have at most 1 control link inneighbor(s) (got 1)"),
    ):
        model.pid_control.add(
            Node(5, Point(0.5, -1.0)),
            [
                pid_control.Static(
                    listen_node_id=[1],
                    target=10.0,
                    proportional=-2.5,
                    integral=-0.001,
                    derivative=10.0,
                )
            ],
        )
        model.link.add(
            model.pid_control[5],
            model.pump[2],
        )
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Node 4 can have at most 1 control link outneighbor(s) (got 1)"
        ),
    ):
        model.pump.add(Node(6, Point(-1.0, 0)), [pump.Static(flow_rate=[0.0])])
        model.link.add(
            model.pid_control[4],
            model.pump[6],
        )


def test_minimum_flow_neighbor():
    model = Model(
        starttime="2020-01-01",
        endtime="2021-01-01",
        crs="EPSG:28992",
        solver=Solver(),
    )

    model.basin.add(
        Node(3, Point(2.0, 0.0)),
        [
            basin.Profile(area=[1000.0, 1000.0], level=[0.0, 1.0]),
            basin.State(level=[0.0]),
        ],
    )
    model.outlet.add(
        Node(2, Point(1.0, 0.0)),
        [outlet.Static(flow_rate=[1e-3], min_upstream_level=[2.0])],
    )
    model.terminal.add(Node(4, Point(3.0, -2.0)))

    with pytest.raises(
        ValueError,
        match=re.escape("Minimum flow inneighbor or outneighbor unsatisfied"),
    ):
        model.link.add(model.basin[3], model.outlet[2])
        model.write("test.toml")


def test_minimum_control_neighbor():
    model = Model(
        starttime="2020-01-01",
        endtime="2021-01-01",
        crs="EPSG:28992",
        solver=Solver(),
    )

    model.basin.add(
        Node(3, Point(2.0, 0.0)),
        [
            basin.Profile(area=[1000.0, 1000.0], level=[0.0, 1.0]),
            basin.State(level=[0.0]),
        ],
    )
    model.outlet.add(
        Node(2, Point(1.0, 0.0)),
        [outlet.Static(flow_rate=[1e-3], min_upstream_level=[2.0])],
    )
    model.terminal.add(Node(4, Point(3.0, -2.0)))

    model.link.add(model.basin[3], model.outlet[2])
    model.link.add(model.outlet[2], model.terminal[4])

    with pytest.raises(
        ValueError,
        match=re.escape("Minimum control inneighbor or outneighbor unsatisfied"),
    ):
        model.pid_control.add(
            Node(5, Point(0.5, 1)),
            [
                pid_control.Static(
                    listen_node_id=[3],
                    target=10.0,
                    proportional=-2.5,
                    integral=-0.001,
                    derivative=10.0,
                )
            ],
        )
        model.write("test.toml")


def test_geometry_validation():
    point = Point(0.0, 0.0)
    poly = point.buffer(1.0)

    assert isinstance(poly, Polygon)
    basinarea = basin.Area(geometry=[poly])
    assert isinstance(basinarea.df.geometry[0], MultiPolygon)

    basinarea = basin.Area(geometry=[basinarea.df.geometry[0]])
    assert isinstance(basinarea.df.geometry[0], MultiPolygon)

    with pytest.raises(ValueError):
        basin.Area(geometry=[point])

    # Drop third dimension on geometry on initialization
    threed = basinarea.df.geometry.force_3d()
    assert threed.has_z.iloc[0]
    basinarea = basin.Area(geometry=threed)
    assert not basinarea.df.geometry.has_z.iloc[0]



# Automatically generated file. Do not modify.

from collections.abc import Callable
from typing import Annotated, Any

import pandas as pd
import pandera as pa
import pyarrow
from pandera.dtypes import Int32
from pandera.typing import Index, Series

from ribasim import migrations


class _BaseSchema(pa.DataFrameModel):
    class Config:
        add_missing_columns = True
        coerce = True

    @classmethod
    def _index_name(self) -> str:
        return "fid"

    @pa.dataframe_parser
    def _name_index(cls, df):
        df.index.name = cls._index_name()
        return df

    @classmethod
    def migrate(cls, df: Any, schema_version: int) -> Any:
        f: Callable[[Any, Any], Any] = getattr(
            migrations, str(cls.__name__).lower() + "_migration", lambda x, _: x
        )
        return f(df, schema_version)


class BasinConcentrationExternalSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    substance: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=False
    )
    concentration: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )


class BasinConcentrationStateSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    substance: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=False
    )
    concentration: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )


class BasinConcentrationSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    substance: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=False
    )
    drainage: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    precipitation: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )


class BasinProfileSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    area: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(nullable=False)
    level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )


class BasinStateSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )


class BasinStaticSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    drainage: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    potential_evaporation: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = (
        pa.Field(nullable=True)
    )
    infiltration: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    precipitation: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )


class BasinSubgridTimeSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    subgrid_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False
    )
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    basin_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    subgrid_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )


class BasinSubgridSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    subgrid_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False
    )
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    basin_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    subgrid_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )


class BasinTimeSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    drainage: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    potential_evaporation: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = (
        pa.Field(nullable=True)
    )
    infiltration: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    precipitation: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )


class ContinuousControlFunctionSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    input: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    output: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    controlled_variable: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=False
    )


class ContinuousControlVariableSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    listen_node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False
    )
    variable: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=False
    )
    weight: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    look_ahead: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )


class DiscreteControlConditionSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    compound_variable_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False
    )
    condition_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False
    )
    greater_than: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=True
    )


class DiscreteControlLogicSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    truth_state: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=False
    )
    control_state: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=False
    )


class DiscreteControlVariableSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    compound_variable_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False
    )
    listen_node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False
    )
    variable: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=False
    )
    weight: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    look_ahead: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )


class FlowBoundaryConcentrationSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    substance: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=False
    )
    concentration: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )


class FlowBoundaryStaticSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    active: Series[Annotated[pd.ArrowDtype, pyarrow.bool_()]] = pa.Field(nullable=True)
    flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )


class FlowBoundaryTimeSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )


class FlowDemandStaticSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    demand: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    demand_priority: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=True
    )


class FlowDemandTimeSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    demand: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    demand_priority: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=True
    )


class LevelBoundaryConcentrationSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    substance: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=False
    )
    concentration: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )


class LevelBoundaryStaticSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    active: Series[Annotated[pd.ArrowDtype, pyarrow.bool_()]] = pa.Field(nullable=True)
    level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )


class LevelBoundaryTimeSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )


class LevelDemandStaticSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    min_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    max_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    demand_priority: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=True
    )


class LevelDemandTimeSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    min_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    max_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    demand_priority: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=True
    )


class LinearResistanceStaticSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    active: Series[Annotated[pd.ArrowDtype, pyarrow.bool_()]] = pa.Field(nullable=True)
    resistance: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    max_flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    control_state: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=True
    )


class ManningResistanceStaticSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    active: Series[Annotated[pd.ArrowDtype, pyarrow.bool_()]] = pa.Field(nullable=True)
    length: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    manning_n: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    profile_width: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    profile_slope: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    control_state: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=True
    )


class OutletStaticSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    active: Series[Annotated[pd.ArrowDtype, pyarrow.bool_()]] = pa.Field(nullable=True)
    flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    min_flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    max_flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    min_upstream_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    max_downstream_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = (
        pa.Field(nullable=True)
    )
    control_state: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=True
    )


class OutletTimeSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    min_flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    max_flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    min_upstream_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    max_downstream_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = (
        pa.Field(nullable=True)
    )


class PidControlStaticSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    active: Series[Annotated[pd.ArrowDtype, pyarrow.bool_()]] = pa.Field(nullable=True)
    listen_node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False
    )
    target: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    proportional: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    integral: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    derivative: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    control_state: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=True
    )


class PidControlTimeSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    listen_node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    target: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    proportional: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    integral: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    derivative: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )


class PumpStaticSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    active: Series[Annotated[pd.ArrowDtype, pyarrow.bool_()]] = pa.Field(nullable=True)
    flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    min_flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    max_flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    min_upstream_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    max_downstream_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = (
        pa.Field(nullable=True)
    )
    control_state: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=True
    )


class PumpTimeSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    min_flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    max_flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    min_upstream_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    max_downstream_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = (
        pa.Field(nullable=True)
    )


class TabulatedRatingCurveStaticSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    active: Series[Annotated[pd.ArrowDtype, pyarrow.bool_()]] = pa.Field(nullable=True)
    level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    max_downstream_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = (
        pa.Field(nullable=True)
    )
    control_state: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=True
    )


class TabulatedRatingCurveTimeSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    flow_rate: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    max_downstream_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = (
        pa.Field(nullable=True)
    )


class UserDemandConcentrationSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    substance: Series[Annotated[pd.ArrowDtype, pyarrow.string()]] = pa.Field(
        nullable=False
    )
    concentration: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )


class UserDemandStaticSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    active: Series[Annotated[pd.ArrowDtype, pyarrow.bool_()]] = pa.Field(nullable=True)
    demand: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=True
    )
    return_factor: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    min_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    demand_priority: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=True
    )


class UserDemandTimeSchema(_BaseSchema):
    fid: Index[Int32] = pa.Field(default=1, check_name=True, coerce=True)
    node_id: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=False, default=0
    )
    time: Series[Annotated[pd.ArrowDtype, pyarrow.timestamp("ms")]] = pa.Field(
        nullable=False
    )
    demand: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    return_factor: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    min_level: Series[Annotated[pd.ArrowDtype, pyarrow.float64()]] = pa.Field(
        nullable=False
    )
    demand_priority: Series[Annotated[pd.ArrowDtype, pyarrow.int32()]] = pa.Field(
        nullable=True
    )



__version__ = "2025.2.0"
# Keep synced write_schema_version in ribasim_qgis/core/geopackage.py
__schema_version__ = 5

import logging

logging.getLogger("datacompy").setLevel(logging.ERROR)

from ribasim.config import Allocation, Logging, Node, Solver  # noqa: E402
from ribasim.geometry.link import LinkTable  # noqa: E402
from ribasim.model import Model  # noqa: E402

__all__ = ["LinkTable", "Allocation", "Logging", "Model", "Solver", "Node"]
